from clearskies.contexts.context import Context
from clearskies.input_outputs import Cli as CliInputOutput


class Cli(Context):
    def __call__(self):  # type: ignore
        return self.execute_application(CliInputOutput())



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Any

import clearskies.parameters_to_properties
import clearskies.typing
from clearskies import configs
from clearskies.columns.string import String

if TYPE_CHECKING:
    from clearskies import Model


class CreatedByUserAgent(String):
    """
    This column will automatically take the user agent from the client and store it in the model upon creation.

    If the user agent isn't available from the context being executed, then you may end up with an error
    (depending on the context).  This is a good thing if you are trying to consistely provide audit information,
    but may be a problem if your model creation needs to happen more flexibly.  Example:

    ```python
    import clearskies


    class MyModel(clearskies.Model):
        backend = clearskies.backends.MemoryBackend()
        id_column_name = "id"

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        user_agent = clearskies.columns.CreatedByUserAgent()


    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Create(
            MyModel,
            writeable_column_names=["name"],
            readable_column_names=["id", "name", "user_agent"],
        ),
        classes=[MyModel],
    )
    wsgi()
    ```

    And if you invoked this:

    ```bash
    $ curl 'http://localhost:8080' -d '{"name":"Bob"}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "a66e5fa9-6377-4d3b-9d50-fc7feaed6d1a",
            "name": "Bob",
            "user_agent": "curl/8.5.0"
        },
        "pagination": {},
        "input_errors": {}
    }
    ```
    """

    """
    Since this column is always populated automatically, it is never directly writeable.
    """
    is_writeable = configs.Boolean(default=False)
    _descriptor_config_map = None

    _allowed_search_operators = ["=", "in", "is not null", "is null", "like"]

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        is_readable: bool = True,
        is_searchable: bool = True,
        is_temporary: bool = False,
        on_change_pre_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_post_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_save_finished: clearskies.typing.action | list[clearskies.typing.action] = [],
    ):
        pass

    def pre_save(self, data: dict[str, Any], model: Model) -> dict[str, Any]:
        if model:
            return data
        input_output = self.di.build("input_output", cache=True)
        data = {**data, self.name: input_output.request_headers.user_agent}
        if self.on_change_pre_save:
            data = self.execute_actions_with_data(self.on_change_pre_save, model, data)
        return data



import inspect
from abc import ABC, abstractmethod
from typing import Any, Callable, Type

import clearskies.column
import clearskies.model
import clearskies.query
from clearskies.autodoc.schema import Schema as AutoDocSchema


class Backend(ABC):
    supports_n_plus_one = False
    can_count = True

    @abstractmethod
    def update(self, id: int | str, data: dict[str, Any], model: clearskies.model.Model) -> dict[str, Any]:
        """Update the record with the given id with the information from the data dictionary."""
        pass

    @abstractmethod
    def create(self, data: dict[str, Any], model: clearskies.model.Model) -> dict[str, Any]:
        """Create a record with the information from the data dictionary."""
        pass

    @abstractmethod
    def delete(self, id: int | str, model: clearskies.model.Model) -> bool:
        """Delete the record with the given id."""
        pass

    @abstractmethod
    def count(self, query: clearskies.query.Query) -> int:
        """Return the number of records which match the given query configuration."""
        pass

    @abstractmethod
    def records(
        self, query: clearskies.query.Query, next_page_data: dict[str, str | int] | None = None
    ) -> list[dict[str, Any]]:
        """
        Return a list of records that match the given query configuration.

        next_page_data is used to return data to the caller.  Pass in an empty dictionary, and it will be populated
        with the data needed to return the next page of results.  If it is still an empty dictionary when returned,
        then there is no additional data.
        """
        pass

    @abstractmethod
    def validate_pagination_data(self, data: dict[str, Any], case_mapping: Callable[[str], str]) -> str:
        """
        Check if the given dictionary is valid pagination data for the background.

        Return a string with an error message, or an empty string if the data is valid
        """
        pass

    @abstractmethod
    def allowed_pagination_keys(self) -> list[str]:
        """
        Return the list of allowed keys in the pagination kwargs for the backend.

        It must always return keys in snake_case so that the auto casing system can
        adjust on the front-end for consistency.
        """
        pass

    @abstractmethod
    def documentation_pagination_next_page_response(self, case_mapping: Callable) -> list[Any]:
        """
        Return a list of autodoc schema objects.

        It will describe the contents of the `next_page` dictionary
        in the pagination section of the response
        """
        pass

    @abstractmethod
    def documentation_pagination_parameters(self, case_mapping: Callable) -> list[tuple[AutoDocSchema, str]]:
        """
        Return a list of autodoc schema objects describing the allowed input keys to set pagination.

        It should return a list of tuples, with each tuple corresponding to an input key.
        The first element in the tuple should be the schema, and the second should be the description.
        """
        pass

    @abstractmethod
    def documentation_pagination_next_page_example(self, case_mapping: Callable) -> dict[str, Any]:
        """
        Return an example for next page documentation.

        Returns an example (as a simple dictionary) of what the next_page data in the pagination response
        should look like
        """
        pass

    def column_from_backend(self, column: clearskies.column.Column, value: Any) -> Any:
        """
        Manage transformations from the backend.

        The idea with this (and `column_to_backend`) is that the transformations to
        and from the backend are mostly determined by the column type - integer, string,
        date, etc...  However, there are cases where these are also backend specific: a datetime
        column may be serialized different ways for different databases, a JSON column must be
        serialized for a database but won't be serialized for an API call, etc...  Therefore
        we mostly just let the column handle this, but we want the backend to be in charge
        in case it needs to make changes.
        """
        return column.from_backend(value)

    def column_to_backend(self, column: clearskies.column.Column, backend_data: dict[str, Any]) -> dict[str, Any]:
        """
        Manage transformations to the backend.

        The idea with this (and `column_from_backend`) is that the transformations to
        and from the backend are mostly determined by the column type - integer, string,
        date, etc...  However, there are cases where these are also backend specific: a datetime
        column may be serialized different ways for different databases, a JSON column must be
        serialized for a database but won't be serialized for an API call, etc...  Therefore
        we mostly just let the column handle this, but we want the backend to be in charge
        in case it needs to make changes.
        """
        return column.to_backend(backend_data)



import clearskies.typing
from clearskies import parameters_to_properties
from clearskies.columns.has_many import HasMany


class HasManySelf(HasMany):
    """
    This is just like the HasMany column, but is used when the model references itself.

    This exists because a model can't refer to itself inside it's own class definition.  There are
    workarounds, but having this class is usually quicker for the developer.

    The main difference between this and HasMany is that you don't have to provide the child class.
    Also, the name of the column that contains the id of the parent becomes `parent_id` by default,
    rather than basing it on the name of the model.  This is done because, since the model is also
    the child, using the name of the model in the name of the column id is often ambiguous.

    See also BelongsToSelf.
    """

    _descriptor_config_map = None

    @parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        foreign_column_name: str | None = None,
        readable_child_columns: list[str] = [],
        where: clearskies.typing.condition | list[clearskies.typing.condition] = [],
        is_readable: bool = True,
        on_change_pre_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_post_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_save_finished: clearskies.typing.action | list[clearskies.typing.action] = [],
    ):
        pass

    def finalize_configuration(self, model_class, name) -> None:
        """
        Finalize and check the configuration.

        This is an external trigger called by the model class when the model class is ready.
        The reason it exists here instead of in the constructor is because some columns are tightly
        connected to the model class, and can't validate configuration until they know what the model is.
        Therefore, we need the model involved, and the only way for a property to know what class it is
        in is if the parent class checks in (which is what happens here).
        """
        self.child_model_class = model_class
        has_value = False
        try:
            has_value = bool(self.foreign_column_name)
        except KeyError:
            pass

        if not has_value:
            self.foreign_column_name = "parent_id"

        super().finalize_configuration(model_class, name)



from __future__ import annotations

from typing import TYPE_CHECKING

from clearskies.configs import config

if TYPE_CHECKING:
    from clearskies.column import Column


class Columns(config.Config):
    """This is for a configuration that should be a dictionary of columns with the key being the column name."""

    def __set__(self, instance, value: dict[str, Column]):
        if value is None:
            return

        if not isinstance(value, dict):
            error_prefix = self._error_prefix(instance)
            raise TypeError(
                f"{error_prefix} attempt to set a value of type '{value.__class__.__name__}' to a dictionary with columns"
            )
        for index, item in enumerate(value.values()):
            if not hasattr(item, "on_change_pre_save"):
                error_prefix = self._error_prefix(instance)
                raise TypeError(
                    f"{error_prefix} attempt to set a value of type '{item.__class__.__name__}' for item #{index + 1}.  A column was expected."
                )
        instance._set_config(self, value)

    def __get__(self, instance, parent) -> dict[str, Column]:
        if not instance:
            return self  # type: ignore
        return instance._get_config(self)



import clearskies.configs
import clearskies.parameters_to_properties
from clearskies.security_header import SecurityHeader


class CacheControl(SecurityHeader):
    max_age = clearskies.configs.Integer()
    s_maxage = clearskies.configs.Integer()
    stale_while_revalidate = clearskies.configs.Integer()
    stale_if_error = clearskies.configs.Integer()
    immutable = clearskies.configs.Boolean(default=False)
    must_understand = clearskies.configs.Boolean(default=False)
    no_cache = clearskies.configs.Boolean(default=False)
    no_store = clearskies.configs.Boolean(default=False)
    no_transform = clearskies.configs.Boolean(default=False)
    private = clearskies.configs.Boolean(default=False)
    public = clearskies.configs.Boolean(default=False)

    numbers: list[str] = [
        "max_age",
        "stale_if_error",
        "stale_while_revalidate",
        "s_maxage",
    ]
    bools: list[str] = [
        "immutable",
        "must_understand",
        "no_cache",
        "no_store",
        "no_transform",
        "private",
        "public",
    ]

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        max_age: int | None = None,
        s_maxage: int | None = None,
        stale_while_revalidate: int | None = None,
        stale_if_error: int | None = None,
        immutable: bool = False,
        must_understand: bool = False,
        no_cache: bool = False,
        no_store: bool = False,
        no_transform: bool = False,
        private: bool = False,
        public: bool = False,
    ):
        self.finalize_and_validate_configuration()

    def set_headers_for_input_output(self, input_output):
        parts = []
        for variable_name in self.bools:
            value = getattr(self, variable_name)
            if not value:
                continue
            parts.append(variable_name.replace("_", "-"))
        for variable_name in self.numbers:
            value = getattr(self, variable_name)
            if value is None:
                continue
            key_name = variable_name.replace("_", "-")
            parts.append(f"{key_name}={value}")
        if not parts:
            return
        input_output.response_headers.add("cache-control", ", ".join(parts))



import json
import urllib
import urllib.parse
from typing import Callable

from .input_output import InputOutput


class Wsgi(InputOutput):
    _environment: dict[str, str] = {}
    _start_response: Callable = None  # type: ignore
    _request_headers: dict[str, str] = {}
    _cached_body: str | None = None

    def __init__(self, environment, start_response):
        self._environment = environment
        self._start_response = start_response
        self._request_headers = {}
        for key, value in self._environment.items():
            if key.upper()[0:5] == "HTTP_":
                self._request_headers[key[5:].lower()] = value
        super().__init__()

    def _from_environment(self, key):
        return self._environment[key] if key in self._environment else ""

    def respond(self, body, status_code=200):
        if "content-type" not in self.response_headers:
            self.response_headers.content_type = "application/json; charset=UTF-8"

        self._start_response(f"{status_code} Ok", [header for header in self.response_headers.items()])  # type: ignore
        if type(body) == bytes:
            final_body = body
        elif type(body) == str:
            final_body = body.encode("utf-8")
        else:
            final_body = json.dumps(body).encode("utf-8")
        return [final_body]

    def has_body(self):
        return bool(self._from_environment("CONTENT_LENGTH"))

    def get_body(self):
        if self._cached_body is None:
            self._cached_body = (
                self._from_environment("wsgi.input").read(int(self._from_environment("CONTENT_LENGTH"))).decode("utf-8")
                if self._from_environment("CONTENT_LENGTH")
                else ""
            )
        return self._cached_body

    def get_request_method(self):
        return self._from_environment("REQUEST_METHOD").upper()

    def get_script_name(self):
        return self._from_environment("SCRIPT_NAME")

    def get_path_info(self):
        return self._from_environment("PATH_INFO")

    def get_query_string(self):
        return self._from_environment("QUERY_STRING")

    def get_content_type(self):
        return self._from_environment("CONTENT_TYPE")

    def get_protocol(self):
        return self._from_environment("wsgi.url_scheme").lower()

    def context_specifics(self):
        return {"wsgi_environment": self._environment}

    def get_client_ip(self):
        return self._environment.get("REMOTE_ADDR")

    def get_request_headers(self):
        return self._request_headers



from __future__ import annotations

from collections import OrderedDict
from typing import TYPE_CHECKING, Any, Callable, Self, overload

import clearskies.parameters_to_properties
import clearskies.typing
from clearskies import configs
from clearskies.autodoc.schema import Array as AutoDocArray
from clearskies.autodoc.schema import Object as AutoDocObject
from clearskies.column import Column
from clearskies.columns.many_to_many_ids import ManyToManyIds
from clearskies.functional import string

if TYPE_CHECKING:
    from clearskies import Model


class ManyToManyModels(Column):
    """
    A companion for the ManyToManyIds column that returns the matching models instead of the ids.

    See the example in the ManyToManyIds column to understand how to use it.
    """

    """ The name of the many-to-many column we are attached to. """
    many_to_many_column_name = configs.ModelColumn(required=True)

    is_writeable = configs.Boolean(default=False)
    is_searchable = configs.Boolean(default=False)
    _descriptor_config_map = None

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        many_to_many_column_name,
    ):
        pass

    def finalize_configuration(self, model_class: type, name: str) -> None:
        """Finalize and check the configuration."""
        getattr(self.__class__, "many_to_many_column_name").set_model_class(model_class)
        self.model_class = model_class
        self.name = name
        self.finalize_and_validate_configuration()

        # finally, make sure we're really pointed at a many-to-many column
        many_to_many_column = getattr(model_class, self.many_to_many_column_name)
        if not isinstance(many_to_many_column, ManyToManyIds):
            raise ValueError(
                f"Error with configuration for {model_class.__name__}.{name}, which is a ManyToManyModels column.  It needs to point to a ManyToManyIds column, and it was told to use {model_class.__name__}.{self.many_to_many_column_name}, but this is not a ManyToManyIds column."
            )

    @property
    def pivot_model(self):
        return self.di.build(self.many_to_many_column.pivot_model_class, cache=True)  # type: ignore

    @property
    def related_models(self):
        return self.di.build(self.many_to_many_column.related_model_class, cache=True)  # type: ignore

    @property
    def related_columns(self):
        return self.related_models.get_columns()

    @property
    def many_to_many_column(self) -> ManyToManyIds:
        return getattr(self.model_class, self.many_to_many_column_name)

    @overload
    def __get__(self, instance: None, cls: type[Model]) -> Self:
        pass

    @overload
    def __get__(self, instance: Model, cls: type[Model]) -> Model:
        pass

    def __get__(self, instance, cls):
        if instance is None:
            self.model_class = cls
            return self

        # this makes sure we're initialized
        if "name" not in self._config:  # type: ignore
            instance.get_columns()

        return self.many_to_many_column.get_related_models(instance)  # type: ignore

    def __set__(self, instance, value: Model | list[Model] | list[dict[str, Any]]) -> None:
        # we allow a list of models or a model, but if it's a model it may represent a single record or a query.
        # if it's a single record then we want to wrap it in a list so we can iterate over it.
        if hasattr(value, "_data") and value._data:
            value = []
        many_to_many_column: ManyToManyIds = self.many_to_many_column  # type: ignore
        related_model_class = many_to_many_column.related_model_class
        related_id_column_name = related_model_class.id_column_name
        record_ids = []
        for index, record in enumerate(value):
            if isinstance(record, dict):
                if not record.get(related_id_column_name):
                    raise KeyError(
                        f"A list of dictionaries was set to '{self.model_class.__name__}.{self.name}', in which case each dictionary should contain the key '{related_id_column_name}', which should be the id of an entry for the '{related_model_class.__name__}' model.  However, no such key was found for entry #{index + 1}"
                    )
                record_ids.append(record[related_id_column_name])
                continue

            # if we get here then the entry should be a model for our related model class
            if not isinstance(record, related_model_class):
                raise TypeError(
                    f"Models were sent to '{self.model_class.__name__}.{self.name}', in which case it should be a list of models of type {related_model_class.__name__}.  However, an object of type '{record.__class__.__name__}' was found for entry #{index + 1}"
                )
            record_ids.append(getattr(record, related_id_column_name))
        setattr(instance, self.many_to_many_column_name, record_ids)

    def add_search(self, model: Model, value: str, operator: str = "", relationship_reference: str = "") -> Model:
        return self.many_to_many_column.add_search(  # type: ignore
            model, value, operator, relationship_reference=relationship_reference
        )  # type: ignore

    def to_json(self, model: Model) -> dict[str, Any]:
        records = []
        many_to_many_column: ManyToManyIds = self.many_to_many_column  # type: ignore
        columns = many_to_many_column.related_columns
        related_id_column_name = many_to_many_column.related_model_class.id_column_name
        for related in many_to_many_column.get_related_models(model):
            json = OrderedDict()
            if related_id_column_name not in many_to_many_column.readable_related_column_names:
                json[related_id_column_name] = columns[related_id_column_name].to_json(related)
            for column_name in many_to_many_column.readable_related_column_names:
                column_data = columns[column_name].to_json(related)
                if type(column_data) == dict:
                    json = {**json, **column_data}  # type: ignore
                else:
                    json[column_name] = column_data
            records.append(json)
        return {self.name: records}

    def documentation(self, name: str | None = None, example: str | None = None, value: str | None = None):
        many_to_many_column = self.many_to_many_column  # type: ignore
        columns = many_to_many_column.related_columns
        related_id_column_name = many_to_many_column.related_model_class.id_column_name
        related_properties = [columns[related_id_column_name].documentation()]

        for column_name in many_to_many_column.readable_related_column_names:
            related_docs = columns[column_name].documentation()
            if type(related_docs) != list:
                related_docs = [related_docs]
            related_properties.extend(related_docs)

        related_object = AutoDocObject(
            string.title_case_to_nice(many_to_many_column.related_model_class.__name__),
            related_properties,
        )
        return AutoDocArray(name if name is not None else self.name, related_object, value=value)



from __future__ import annotations

import re
from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Callable, Iterator, Self

from clearskies.autodoc.schema import Schema as AutoDocSchema
from clearskies.di import InjectableProperties, inject
from clearskies.functional import string
from clearskies.query import Condition, Join, Query, Sort
from clearskies.schema import Schema

if TYPE_CHECKING:
    from clearskies import Column
    from clearskies.backends import Backend


class Model(Schema, InjectableProperties):
    """
    A clearskies model.

    To be useable, a model class needs three things:

     1. Column definitions
     2. The name of the id column
     3. A backend
     4. A destination name (equivalent to a table name for SQL backends)

    """

    _previous_data: dict[str, Any] = {}
    _data: dict[str, Any] = {}
    _next_data: dict[str, Any] = {}
    _transformed_data: dict[str, Any] = {}
    _touched_columns: dict[str, bool] = {}
    _query: Query | None = None
    _query_executed: bool = False
    _count: int | None = None
    _next_page_data: dict[str, Any] | None = None

    id_column_name: str = ""
    backend: Backend = None  # type: ignore

    _di = inject.Di()

    def __init__(self):
        if not self.id_column_name:
            raise ValueError(
                f"You must define the 'id_column_name' property for every model class, but this is missing for model '{self.__class__.__name__}'"
            )
        if not isinstance(self.id_column_name, str):
            raise TypeError(
                f"The 'id_column_name' property of a model must be a string that specifies the name of the id column, but that is not the case for model '{self.__class__.__name__}'."
            )
        if not self.backend:
            raise ValueError(
                f"You must define the 'backend' property for every model class, but this is missing for model '{self.__class__.__name__}'"
            )
        if not hasattr(self.backend, "documentation_pagination_parameters"):
            raise TypeError(
                f"The 'backend' property of a model must be an object that extends the clearskies.Backend class, but that is not the case for model '{self.__class__.__name__}'."
            )
        self._previous_data = {}
        self._data = {}
        self._next_data = {}
        self._transformed_data = {}
        self._touched_columns = {}
        self._query = None
        self._query_executed = False
        self._count = None
        self._next_page_data = None

    @classmethod
    def destination_name(cls: type[Self]) -> str:
        """
        Return the name of the destination that the model uses for data storage.

        For SQL backends, this would return the table name.  Other backends will use this
        same function but interpret it in whatever way it makes sense.  For instance, an
        API backend may treat it as a URL (or URL path), an SQS backend may expect a queue
        URL, etc...

        By default this takes the class name, converts from title case to snake case, and then
        makes it plural.
        """
        singular = string.camel_case_to_snake_case(cls.__name__)
        if singular[-1] == "y":
            return singular[:-1] + "ies"
        if singular[-1] == "s":
            return singular + "es"
        return f"{singular}s"

    def supports_n_plus_one(self: Self):
        return self.backend.supports_n_plus_one  #  type: ignore

    def __bool__(self: Self) -> bool:  # noqa: D105
        if self._query:
            return bool(self.__len__())

        return True if self._data else False

    def get_raw_data(self: Self) -> dict[str, Any]:
        self.no_queries()
        return self._data

    def set_raw_data(self: Self, data: dict[str, Any]) -> None:
        self.no_queries()
        self._data = {} if data is None else data
        self._transformed_data = {}

    def save(self: Self, data: dict[str, Any] | None = None, columns: dict[str, Column] = {}, no_data=False) -> bool:
        """
        Save data to the database and update the model.

        Executes an update if the model corresponds to a record already, or an insert if not.

        There are two supported flows.  One is to pass in a dictionary of data to save:

        ```python
        model.save({
            "some_column": "New Value",
            "another_column": 5,
        })
        ```

        And the other is to set new values on the columns attributes and then call save without data:

        ```python
        model.some_column = "New Value"
        model.another_column = 5
        model.save()
        ```

        You cannot combine these methods.  If you set a value on a column attribute and also pass
        in a dictionary of data to the save, then an exception will be raised.
        """
        self.no_queries()
        if not data and not self._next_data and not no_data:
            raise ValueError("You have to pass in something to save, or set no_data=True in your call to save/create.")
        if data and self._next_data:
            raise ValueError(
                "Save data was provided to the model class by both passing in a dictionary and setting new values on the column attributes.  This is not allowed.  You will have to use just one method of specifying save data."
            )
        if not data:
            data = {**self._next_data}
            self._next_data = {}

        save_columns = self.get_columns()
        if columns is not None:
            for column in columns.values():
                save_columns[column.name] = column

        old_data = self.get_raw_data()
        data = self.columns_pre_save(data, save_columns)
        data = self.pre_save(data)
        if data is None:
            raise ValueError("pre_save forgot to return the data array!")

        [to_save, temporary_data] = self.columns_to_backend(data, save_columns)
        to_save = self.to_backend(to_save, save_columns)
        if self:
            new_data = self.backend.update(self._data[self.id_column_name], to_save, self)  # type: ignore
        else:
            new_data = self.backend.create(to_save, self)  # type: ignore
        id = self.backend.column_from_backend(save_columns[self.id_column_name], new_data[self.id_column_name])  # type: ignore

        # if we had any temporary columns add them back in
        new_data = {
            **temporary_data,
            **new_data,
        }

        data = self.columns_post_save(data, id, save_columns)
        self.post_save(data, id)

        self.set_raw_data(new_data)
        self._transformed_data = {}
        self._previous_data = old_data
        self._touched_columns = {key: True for key in data.keys()}

        self.columns_save_finished(save_columns)
        self.save_finished()

        return True

    def is_changing(self: Self, key: str, data: dict[str, Any]) -> bool:
        """
        Return True/False to denote if the given column is being modified by the active save operation.

        Pass in the name of the column to check and the data dictionary from the save in progress
        """
        self.no_queries()
        has_old_value = key in self._data
        has_new_value = key in data

        if not has_new_value:
            return False
        if not has_old_value:
            return True

        return getattr(self, key) != data[key]

    def latest(self: Self, key: str, data: dict[str, Any]) -> Any:
        """
        Return the 'latest' value for a column during the save operation.

        Return either the column value from the data dictionary or the current value stored in the model
        Basically, shorthand for the optimized version of:  `data.get(key, default=getattr(self, key))` (which is
        less than ideal because it always builds the default value, even when not necessary)

        Pass in the name of the column to check and the data dictionary from the save in progress
        """
        self.no_queries()
        if key in data:
            return data[key]
        return getattr(self, key)

    def was_changed(self: Self, key: str) -> bool:
        """Return True/False to denote if a column was changed in the last save."""
        self.no_queries()
        if self._previous_data is None:
            raise ValueError("was_changed was called before a save was finished - you must save something first")
        if key not in self._touched_columns:
            return False

        has_old_value = bool(self._previous_data.get(key))
        has_new_value = bool(self._data.get(key))

        if has_new_value != has_old_value:
            return True

        if not has_old_value:
            return False

        columns = self.get_columns()
        new_value = self._data[key]
        old_value = self._previous_data[key]
        if key not in columns:
            return old_value != new_value
        return not columns[key].values_match(old_value, new_value)

    def previous_value(self: Self, key: str):
        self.no_queries()
        return getattr(self.__class__, key).transform(self._previous_data.get(key))

    def delete(self: Self, except_if_not_exists=True) -> bool:
        self.no_queries()
        if not self:
            if except_if_not_exists:
                raise ValueError("Cannot delete model that already exists")
            return True

        columns = self.get_columns()
        self.columns_pre_delete(columns)
        self.pre_delete()

        self.backend.delete(self._data[self.id_column_name], self)  # type: ignore

        self.columns_post_delete(columns)
        self.post_delete()
        return True

    def columns_pre_save(self: Self, data: dict[str, Any], columns) -> dict[str, Any]:
        """Use the column information present in the model to make any necessary changes before saving."""
        iterate = True
        changed = {}
        while iterate:
            iterate = False
            for column in columns.values():
                data = column.pre_save(data, self)
                if data is None:
                    raise ValueError(
                        f"Column {column.name} of type {column.__class__.__name__} did not return any data for pre_save"
                    )

                # if we have newly chnaged data then we want to loop through the pre-saves again
                if data and column.name not in changed:
                    changed[column.name] = True
                    iterate = True
        return data

    def columns_to_backend(self: Self, data: dict[str, Any], columns) -> Any:
        backend_data = {**data}
        temporary_data = {}
        for column in columns.values():
            if column.is_temporary:
                if column.name in backend_data:
                    temporary_data[column.name] = backend_data[column.name]
                    del backend_data[column.name]
                continue

            backend_data = self.backend.column_to_backend(column, backend_data)  # type: ignore
            if backend_data is None:
                raise ValueError(
                    f"Column {column.name} of type {column.__class__.__name__} did not return any data for to_database"
                )

        return [backend_data, temporary_data]

    def to_backend(self: Self, data: dict[str, Any], columns) -> dict[str, Any]:
        return data

    def columns_post_save(self: Self, data: dict[str, Any], id: str | int, columns) -> dict[str, Any]:
        """Use the column information present in the model to make additional changes as needed after saving."""
        for column in columns.values():
            column.post_save(data, self, id)
        return data

    def columns_save_finished(self: Self, columns) -> None:
        """Call the save_finished method on all of our columns."""
        for column in columns.values():
            column.save_finished(self)

    def post_save(self: Self, data: dict[str, Any], id: str | int) -> None:
        """
        Create a hook to extend so you can provide additional pre-save logic as needed.

        It is passed in the data being saved as well as the id.  It should take action as needed and then return
        either the original data array or an adjusted one if appropriate.
        """
        pass

    def pre_save(self: Self, data: dict[str, Any]) -> dict[str, Any]:
        """
        Create a hook to extend so you can provide additional pre-save logic as needed.

        It is passed in the data being saved and it should return the same data with adjustments as needed
        """
        return data

    def save_finished(self: Self) -> None:
        """
        Create a hook to extend so you can provide additional logic after a save operation has fully completed.

        It has no retrun value and is passed no data.  By the time this fires the model has already been
        updated with the new data.  You can decide on the necessary actions using the `was_changed` and
        the `previous_value` functions.
        """
        pass

    def columns_pre_delete(self: Self, columns: dict[str, Column]) -> None:
        """Use the column information present in the model to make any necessary changes before deleting."""
        for column in columns.values():
            column.pre_delete(self)

    def pre_delete(self: Self) -> None:
        """Create a hook to extend so you can provide additional pre-delete logic as needed."""
        pass

    def columns_post_delete(self: Self, columns: dict[str, Column]) -> None:
        """Use the column information present in the model to make any necessary changes after deleting."""
        for column in columns.values():
            column.post_delete(self)

    def post_delete(self: Self) -> None:
        """Create a hook to extend so you can provide additional post-delete logic as needed."""
        pass

    def where_for_request(
        self: Self,
        models: Self,
        routing_data: dict[str, str],
        authorization_data: dict[str, Any],
        input_output: Any,
        overrides: dict[str, Column] = {},
    ) -> Self:
        """Create a hook to automatically apply filtering whenever the model makes an appearance in a get/update/list/search handler."""
        for column in self.get_columns(overrides=overrides).values():
            models = column.where_for_request(models, routing_data, authorization_data, input_output)  # type: ignore
        return models

    ##############################################################
    ### From here down is functionality related to list/search ###
    ##############################################################
    def has_query(self) -> bool:
        """Whether or not this model instance represents a query."""
        return bool(self._query)

    def get_query(self) -> Query:
        """Fetch the query object in the model."""
        return self._query if self._query else Query(self.__class__)

    def as_query(self) -> Self:
        """
        Make the model queryable.

        This is used to remove the ambiguity of attempting execute a query against a model object that stores a record.

        The reason this exists is because the model class is used both to query as well as to operate on single records, which can cause
        subtle bugs if a developer accidentally confuses the two usages.  Consider the following (partial) example:

        ```python
        def some_function(models):
            model = models.find("id=5")
            if model:
                models.save({"test": "example"})
            other_record = model.find("id=6")
        ```

        In the above example it seems likely that the intention was to use `model.save()`, not `models.save()`.  Similarly, the last line
        should be `models.find()`, not `model.find()`.  To minimize these kinds of issues, clearskies won't let you execute a query against
        an individual model record, nor will it let you execute a save against a model being used to make a query.  In both cases, you'll
        get an exception from clearskies, as the models track exactly how they are being used.

        In some rare cases though, you may want to start a new query aginst a model that represents a single record.  This is most common
        if you have a function that was passed an individual model, and you'd like to use it to fetch more records without having to
        inject the model class more generally.  That's where the `as_query()` method comes in.  It's basically just a way of telling clearskies
        "yes, I really do want to start a query using a model that represents a record".  So, for example:

        ```python
        def some_function(models):
            model = models.find("id=5")
            more_models = model.where("test=example")  # throws an exception.
            more_models = model.as_query().where("test=example")  # works as expected.
        ```
        """
        new_model = self._di.build(self.__class__, cache=False)
        new_model.set_query(Query(self.__class__))
        return new_model

    def set_query(self, query: Query) -> Self:
        """Set the query object."""
        self._query = query
        self._query_executed = False
        return self

    def with_query(self, query: Query) -> Self:
        return self._di.build(self.__class__, cache=False).set_query(query)

    def select(self: Self, select: str) -> Self:
        """
        Add some additional columns to the select part of the query.

        This method returns a new object with the updated query.  The original model object is unmodified.
        Multiple calls to this method add together.  The following:

        ```python
        models.select("column_1 column_2").select("column_3")
        ```

        will select column_1, column_2, column_3 in the final query.
        """
        self.no_single_model()
        return self.with_query(self.get_query().add_select(select))

    def select_all(self: Self, select_all=True) -> Self:
        """
        Set whether or not to select all columns with the query.

        This method returns a new object with the updated query.  The original model object is unmodified.
        """
        self.no_single_model()
        return self.with_query(self.get_query().set_select_all(select_all))

    def where(self: Self, where: str | Condition) -> Self:
        """
        Add the given condition to the query.

        This method returns a new object with the updated query.  The original model object is unmodified.

        Conditions should be an SQL-like string of the form [column][operator][value] with an optional table prefix.
        You can safely inject user input into the value.  The column name will also be checked against the searchable
        columns for the model class, and an exception will be thrown if the column doesn't exist or is not searchable.

        Multiple conditions are always joined with AND.  There is no explicit option for OR.  The closest is using an
        IN condition.

        Examples:
        ```python
        for record in (
            models.where("order_id=5").where("status IN ('ACTIVE','PENDING')").where("other_table.id=asdf")
        ):
            print(record.id)
        ```
        """
        self.no_single_model()
        return self.with_query(self.get_query().add_where(where if isinstance(where, Condition) else Condition(where)))

    def join(self: Self, join: str) -> Self:
        """Add a join clause to the query."""
        self.no_single_model()
        return self.with_query(self.get_query().add_join(Join(join)))

    def is_joined(self: Self, table_name: str, alias: str = "") -> bool:
        """
        Check if a given table was already joined.

        If you provide an alias then it will also verify if the table was joined with the specific alias name.
        """
        for join in self.get_query().joins:
            if join.unaliased_table_name != table_name:
                continue

            if alias and join.alias != alias:
                continue

            return True
        return False

    def group_by(self: Self, group_by_column_name: str) -> Self:
        self.no_single_model()
        return self.with_query(self.get_query().set_group_by(group_by_column_name))

    def sort_by(
        self: Self,
        primary_column_name: str,
        primary_direction: str,
        primary_table_name: str = "",
        secondary_column_name: str = "",
        secondary_direction: str = "",
        secondary_table_name: str = "",
    ) -> Self:
        self.no_single_model()
        sort = Sort(primary_table_name, primary_column_name, primary_direction)
        secondary_sort = None
        if secondary_column_name and secondary_direction:
            secondary_sort = Sort(secondary_table_name, secondary_column_name, secondary_direction)
        return self.with_query(self.get_query().set_sort(sort, secondary_sort))

    def limit(self: Self, limit: int) -> Self:
        self.no_single_model()
        return self.with_query(self.get_query().set_limit(limit))

    def pagination(self: Self, **pagination_data) -> Self:
        self.no_single_model()
        error = self.backend.validate_pagination_data(pagination_data, str)
        if error:
            raise ValueError(
                f"Invalid pagination data for model {self.__class__.__name__} with backend "
                + f"{self.backend.__class__.__name__}. {error}"
            )
        return self.with_query(self.get_query().set_pagination(pagination_data))

    def find(self: Self, where: str | Condition) -> Self:
        """
        Return the first model matching a given where condition.

        This is just shorthand for `models.where("column=value").find()`.  Example:

        ```python
        model = models.find("column=value")
        print(model.id)
        ```
        """
        self.no_single_model()
        return self.where(where).first()

    def __len__(self: Self):  # noqa: D105
        self.no_single_model()
        if self._count is None:
            self._count = self.backend.count(self.get_query())
        return self._count

    def __iter__(self: Self) -> Iterator[Self]:  # noqa: D105
        self.no_single_model()
        self._next_page_data = {}
        raw_rows = self.backend.records(
            self.get_query(),
            next_page_data=self._next_page_data,
        )
        return iter([self.model(row) for row in raw_rows])

    def paginate_all(self: Self) -> list[Self]:
        """
        Loop through all available pages of results and returns a list of all models that match the query.

        NOTE: this loads up all records in memory before returning (e.g. it isn't using generators yet), so
        expect delays for large record sets.

        ```python
        for model in models.where("column=value").paginate_all():
            print(model.id)
        ```
        """
        self.no_single_model()
        next_models = self.with_query(self.get_query())
        results = list(next_models.__iter__())
        next_page_data = next_models.next_page_data()
        while next_page_data:
            next_models = self.pagination(**next_page_data)
            results.extend(next_models.__iter__())
            next_page_data = next_models.next_page_data()
        return results

    def model(self: Self, data: dict[str, Any] = {}) -> Self:
        """
        Create a new model object and populates it with the data in `data`.

        NOTE: the difference between this and `model.create` is that model.create() actually saves a record in the backend,
        while this method just creates a model object populated with the given data.
        """
        model = self._di.build(self.__class__, cache=False)
        model.set_raw_data(data)
        return model

    def empty(self: Self) -> Self:
        """
        An alias for self.model({})
        """
        return self.model({})

    def create(self: Self, data: dict[str, Any] = {}, columns: dict[str, Column] = {}, no_data=False) -> Self:
        """
        Create a new record in the backend using the information in `data`.

        new_model = models.create({"column": "value"})
        """
        empty = self.model()
        empty.save(data, columns=columns, no_data=no_data)
        return empty

    def first(self: Self) -> Self:
        """
        Return the first model matching the given query.

        ```python
        model = models.where("column=value").sort_by("age", "DESC").first()
        print(model.id)
        ```
        """
        self.no_single_model()
        iter = self.__iter__()
        try:
            return iter.__next__()
        except StopIteration:
            return self.model()

    def allowed_pagination_keys(self: Self) -> list[str]:
        return self.backend.allowed_pagination_keys()

    def validate_pagination_data(self, kwargs: dict[str, Any], case_mapping: Callable[[str], str]) -> str:
        return self.backend.validate_pagination_data(kwargs, case_mapping)

    def next_page_data(self: Self):
        return self._next_page_data

    def documentation_pagination_next_page_response(self: Self, case_mapping: Callable) -> list[Any]:
        return self.backend.documentation_pagination_next_page_response(case_mapping)

    def documentation_pagination_next_page_example(self: Self, case_mapping: Callable) -> dict[str, Any]:
        return self.backend.documentation_pagination_next_page_example(case_mapping)

    def documentation_pagination_parameters(self: Self, case_mapping: Callable) -> list[tuple[AutoDocSchema, str]]:
        return self.backend.documentation_pagination_parameters(case_mapping)

    def no_queries(self) -> None:
        if self._query:
            raise ValueError(
                "You attempted to save/read record data for a model being used to make a query.  This is not allowed, as it is typically a sign of a bug in your application code."
            )

    def no_single_model(self):
        if self._data:
            raise ValueError(
                "You have attempted to execute a query against a model that represents an individual record.  This is not allowed, as it is typically a sign of a bug in your application code.  If this is intentional, call model.as_query() before executing your query."
            )


class ModelClassReference:
    @abstractmethod
    def get_model_class(self) -> type[Model]:
        pass



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Any

import clearskies.di
import clearskies.parameters_to_properties
import clearskies.typing
from clearskies import configs
from clearskies.columns.datetime import Datetime

if TYPE_CHECKING:
    from clearskies import Model


class Updated(Datetime):
    """
    The updated column records the time that a record is created or updated.

    Note that this will always populate the column anytime the model is created or updated.
    You don't have to provide the timestamp yourself and you should never expose it as
    a writeable column through an endpoint (in fact, you can't).

    ```python
    import clearskies
    import time


    class MyModel(clearskies.Model):
        backend = clearskies.backends.MemoryBackend()
        id_column_name = "id"

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        created = clearskies.columns.Created()
        updated = clearskies.columns.Updated()


    def test_updated(my_models: MyModel) -> MyModel:
        my_model = my_models.create({"name": "Jane"})
        updated_column_after_create = my_model.updated

        time.sleep(2)

        my_model.save({"name": "Susan"})

        return {
            "updated_column_after_create": updated_column_after_create.isoformat(),
            "updated_column_at_end": my_model.updated.isoformat(),
            "difference_in_seconds": (my_model.updated - updated_column_after_create).total_seconds(),
        }


    cli = clearskies.contexts.Cli(clearskies.endpoints.Callable(test_updated), classes=[MyModel])
    cli()
    ```

    And when invoked:

    ```bash
    $ ./test.py | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "updated_column_after_create": "2025-05-18T19:28:46+00:00",
            "updated_column_at_end": "2025-05-18T19:28:48+00:00",
            "difference_in_seconds": 2.0
        },
        "pagination": {},
        "input_errors": {}
    }
    ```

    Note that the `updated` column was set both when the record was first created and when it was updated,
    so there is a two second difference between them (since we slept for two seconds).

    """

    """
    Created fields are never writeable because they always set the created time automatically.
    """
    is_writeable = configs.Boolean(default=False)
    _descriptor_config_map = None

    now = clearskies.di.inject.Now()

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        in_utc: bool = True,
        date_format: str = "%Y-%m-%d %H:%M:%S",
        backend_default: str = "0000-00-00 00:00:00",
        is_readable: bool = True,
        is_searchable: bool = True,
        is_temporary: bool = False,
        on_change_pre_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_post_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_save_finished: clearskies.typing.action | list[clearskies.typing.action] = [],
    ):
        pass

    def pre_save(self, data: dict[str, Any], model: Model) -> dict[str, Any]:
        now = self.now
        if self.timezone_aware:
            now = now.astimezone(self.timezone)
        data = {**data, self.name: now}
        if self.on_change_pre_save:
            data = self.execute_actions_with_data(self.on_change_pre_save, model, data)
        return data



import datetime
from types import ModuleType
from typing import Any, Callable
from wsgiref.simple_server import make_server
from wsgiref.util import setup_testing_defaults

import clearskies.endpoint
import clearskies.endpoint_group
from clearskies.contexts.context import Context
from clearskies.di import AdditionalConfig
from clearskies.input_outputs import Wsgi as WsgiInputOutput


class WsgiRef(Context):
    port: int = 8080

    def __init__(
        self,
        application: Callable | clearskies.endpoint.Endpoint | clearskies.endpoint_group.EndpointGroup,
        port: int = 8080,
        classes: type | list[type] = [],
        modules: ModuleType | list[ModuleType] = [],
        bindings: dict[str, Any] = {},
        additional_configs: AdditionalConfig | list[AdditionalConfig] = [],
        class_overrides: dict[type, type] = {},
        overrides: dict[str, type] = {},
        now: datetime.datetime | None = None,
        utcnow: datetime.datetime | None = None,
    ):
        super().__init__(
            application,
            classes=classes,
            modules=modules,
            bindings=bindings,
            additional_configs=additional_configs,
            class_overrides=class_overrides,
            overrides=overrides,
            now=now,
            utcnow=utcnow,
        )
        self.port = port

    def __call__(self):  # type: ignore
        with make_server("", self.port, self.handler) as httpd:
            print(f"Starting WSGI server on port {self.port}.  This is NOT intended for production usage.")
            httpd.serve_forever()

    def handler(self, environment, start_response):
        return self.execute_application(WsgiInputOutput(environment, start_response))



from clearskies.configs import config


class SelectList(config.Config):
    """
    This is for a configuration that should be a list of strings matching some list of allowed values.

    The allowed values are set when you create the config, and when values are set for the config
    they must match.

    This is different than StringList, which also accepts a list of any strings.
    """

    def __init__(self, allowed_values: list[str], required=False, default=None):
        self.allowed_values = allowed_values
        self.required = required
        self.default = default

    def __set__(self, instance, value: list[str]):
        if value is None:
            return

        if not isinstance(value, list):
            error_prefix = self._error_prefix(instance)
            raise TypeError(
                f"{error_prefix} attempt to set a value of type '{value.__class__.__name__}' to a list parameter"
            )
        for index, item in enumerate(value):
            if not isinstance(item, str):
                error_prefix = self._error_prefix(instance)
                raise TypeError(
                    f"{error_prefix} attempt to set a value of type '{value.__class__.__name__}' for item #{index + 1}.  A string was expected."
                )

            if item not in self.allowed_values:
                error_prefix = self._error_prefix(instance)
                raise ValueError(
                    f"{error_prefix} attempt to set a value of '{item}' for item #{index + 1}.  This is not in the list of allowed values.  It must be one of '"
                    + "', '".join(self.allowed_values)
                    + "'"
                )
        instance._set_config(self, [*value])

    def __get__(self, instance, parent) -> list[str]:
        if not instance:
            return self  # type: ignore
        return instance._get_config(self)



from __future__ import annotations

from collections import OrderedDict
from typing import TYPE_CHECKING, Any, Callable, Self, overload

import clearskies.typing
from clearskies import configs, parameters_to_properties  # type: ignore
from clearskies.autodoc.schema import Array as AutoDocArray
from clearskies.autodoc.schema import Object as AutoDocObject
from clearskies.column import Column
from clearskies.columns.many_to_many_ids import ManyToManyIds
from clearskies.functional import string

if TYPE_CHECKING:
    from clearskies import Model


class ManyToManyPivots(Column):
    """
    A companion for the ManyToManyIds column that returns the matching pivot models instead of the ids.

    See ManyToManyIdsWithData for an example of how to use it (but note that it works just the same for the
    ManyToManyIds column).
    """

    """ The name of the many-to-many column we are attached to. """
    many_to_many_column_name = configs.ModelColumn(required=True)

    is_writeable = configs.Boolean(default=False)
    is_searchable = configs.Boolean(default=False)
    _descriptor_config_map = None

    @parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        many_to_many_column_name,
    ):
        pass

    def finalize_configuration(self, model_class: type, name: str) -> None:
        """Finalize and check the configuration."""
        getattr(self.__class__, "many_to_many_column_name").set_model_class(model_class)
        self.model_class = model_class
        self.name = name
        self.finalize_and_validate_configuration()

        # finally, make sure we're really pointed at a many-to-many column
        many_to_many_column = getattr(model_class, self.many_to_many_column_name)
        if not isinstance(many_to_many_column, ManyToManyIds):
            raise ValueError(
                f"Error with configuration for {model_class.__name__}.{name}, which is a ManyToManyModels column.  It needs to point to a ManyToManyIds column, and it was told to use {model_class.__name__}.{self.many_to_many_column_name}, but this is not a ManyToManyIds column."
            )

    @property
    def pivot_model(self) -> Model:
        return self.di.build(self.many_to_many_column.pivot_model_class, cache=True)  # type: ignore

    @property
    def related_models(self) -> Model:
        return self.di.build(self.many_to_many_column.related_model_class, cache=True)  # type: ignore

    @property
    def related_columns(self):
        return self.related_models.get_columns()

    @property
    def many_to_many_column(self) -> ManyToManyIds:
        return getattr(self.model_class, self.many_to_many_column_name)

    @overload
    def __get__(self, instance: None, cls: type[Model]) -> Self:
        pass

    @overload
    def __get__(self, instance: Model, cls: type[Model]) -> Model:
        pass

    def __get__(self, instance, cls):
        if instance is None:
            self.model_class = cls
            return self

        # this makes sure we're initialized
        if "name" not in self._config:  # type: ignore
            instance.get_columns()

        many_to_many_column = self.many_to_many_column  # type: ignore
        own_column_name_in_pivot = many_to_many_column._config("own_column_name_in_pivot")
        my_id = getattr(instance, instance.id_column_name)
        return [model for model in self.pivot_model.where(f"{own_column_name_in_pivot}={my_id}")]

    def __set__(self, instance, value: Model | list[Model] | list[dict[str, Any]]) -> None:
        raise NotImplementedError("Saving not supported for ManyToManyPivots")

    def add_search(self, model: Model, value: str, operator: str = "", relationship_reference: str = "") -> Model:
        raise NotImplementedError("Searching not supported for ManyToManyPivots")

    def to_json(self, model: Model) -> dict[str, Any]:
        records = []
        many_to_many_column = self.many_to_many_column  # type: ignore
        columns = many_to_many_column.pivot_columns
        readable_column_names = many_to_many_column.readable_pivot_column_names
        pivot_id_column_name = many_to_many_column.pivot_model_class.id_column_name
        for pivot in many_to_many_column.get_pivot_models(model):
            json = OrderedDict()
            if pivot_id_column_name not in readable_column_names:
                json[pivot_id_column_name] = columns[pivot_id_column_name].to_json(pivot)
            for column_name in readable_column_names:
                column_data = columns[column_name].to_json(pivot)
                if type(column_data) == dict:
                    json = {**json, **column_data}  # type: ignore
                else:
                    json[column_name] = column_data
            records.append(json)
        return {self.name: records}

    def documentation(self, name: str | None = None, example: str | None = None, value: str | None = None):
        many_to_many_column = self.many_to_many_column  # type: ignore
        columns = many_to_many_column.pivot_columns
        pivot_id_column_name = many_to_many_column.pivot_model_class.id_column_name
        pivot_properties = [columns[pivot_id_column_name].documentation()]

        for column_name in many_to_many_column.readable_pivot_column_names:
            pivot_docs = columns[column_name].documentation()
            if type(pivot_docs) != list:
                pivot_docs = [pivot_docs]
            pivot_properties.extend(pivot_docs)

        pivot_object = AutoDocObject(
            string.title_case_to_nice(many_to_many_column.pivot_model_class.__name__),
            pivot_properties,
        )
        return AutoDocArray(name if name is not None else self.name, pivot_object, value=value)



from clearskies.configs import select_list


class ModelColumns(select_list.SelectList):
    def __init__(self, model_column_config_name="", allow_relationship_references=False, required=False, default=None):
        self.required = required
        self.default = default
        self.model_column_config_name = model_column_config_name
        self.allow_relationship_references = allow_relationship_references

    def __set__(self, instance, value: list[str]):
        if value is None:
            return

        if not isinstance(value, list):
            error_prefix = self._error_prefix(instance)
            raise TypeError(
                f"{error_prefix} attempt to set a value of type '{value.__class__.__name__}' to a list parameter"
            )

        instance._set_config(self, value)
        # unlike select_list, we won't validate the value currently because we won't be able to
        # do that until the finalize_and_validate_configuration phase

    def get_allowed_columns(self, model_class, column_configs):
        return [name for name in column_configs.keys()]

    def my_description(self):
        return "column"

    def finalize_and_validate_configuration(self, instance):
        super().finalize_and_validate_configuration(instance)

        # when we were created we were told the name of the config that stores
        # the model class we're getting readable columns from.  Let's use that
        # and fetch the model class.  Note that this is an optional feature though
        if not self.model_column_config_name:
            return

        model_class = getattr(instance, self.model_column_config_name)
        values = instance._get_config(self)
        if not values or not model_class:
            return

        allowed_columns = self.get_allowed_columns(model_class, model_class.get_columns())
        for value in values:
            if self.allow_relationship_references and "." in value:
                value = value.split(".")[0]
            if value not in allowed_columns:
                error_prefix = self._error_prefix(instance)
                my_description = self.my_description()
                raise ValueError(
                    f"{error_prefix} attempt to set a value of '{value}' but this is not a {my_description} in the specified model class, '{model_class.__name__}.  Expected values are: '"
                    + "', '".join(allowed_columns)
                    + "'"
                )



import datetime

from clearskies.validators.timedelta import Timedelta


class InTheFutureAtLeast(Timedelta):
    def check_timedelta(self, as_date: datetime.datetime, column_name: str) -> str:
        if as_date < self.utcnow + self.timedelta:
            human_friendly = self.delta_human_friendly()
            return f"'{column_name}' must be at least {human_friendly} in the future."
        return ""



class Response:
    status = None
    schema = None
    description = None

    def __init__(self, status, schema, description=None):
        self.status = status
        self.schema = schema
        self.description = description if description is not None else ""



from clearskies.columns.has_many import HasMany


class HasOne(HasMany):
    """
    This operates exactly like the HasMany relationship, except it assumes there is only ever one child.

    The only real difference between this and HasMany is that the HasMany column type will return a list
    of models, while this returns the first model.
    """

    _descriptor_config_map = None

    pass



from clearskies.authentication.authentication import Authentication
from clearskies.authentication.authorization import Authorization
from clearskies.authentication.authorization_pass_through import AuthorizationPassThrough
from clearskies.authentication.jwks import Jwks
from clearskies.authentication.public import Public
from clearskies.authentication.secret_bearer import SecretBearer

__all__ = [
    "Authentication",
    "Authorization",
    "AuthorizationPassThrough",
    "Jwks",
    "Public",
    "SecretBearer",
]



from .header import Header
from .json_body import JSONBody
from .parameter import Parameter
from .request import Request
from .url_parameter import URLParameter
from .url_path import URLPath

__all__ = [
    "Header",
    "JSONBody",
    "Parameter",
    "Request",
    "URLParameter",
    "URLPath",
]



from typing import Any


class Parameter:
    name: str = ""
    parameter: Any = None
    required: bool = False
    location_map = {
        "url_parameter": "query",
        "header": "header",
        "url_path": "path",
    }

    def __init__(self, oai3_schema_resolver):
        self.oai3_schema_resolver = oai3_schema_resolver

    def set_parameter(self, parameter):
        self.parameter = parameter
        self.name = self.parameter.definition.name
        self.required = self.parameter.required

    def convert(self):
        if self.parameter.location not in self.location_map:
            raise ValueError(
                f"Parameter of class {self.parameter.__class__} declares "
                + f"an unsupported location: '{self.parameter.location}'"
            )

        return {
            "name": self.parameter.definition.name,
            "description": self.parameter.description,
            "required": self.required,
            "in": self.location_map[self.parameter.location],
            "schema": self.oai3_schema_resolver(self.parameter.definition).convert(),
        }



class Config:
    def __init__(self, required=False, default=None):
        self.required = required
        self.default = default

    def _error_prefix(self, instance) -> str:
        name = instance._descriptor_to_name(self)
        class_name = instance.__class__.__name__
        return f"Error with '{class_name}.{name}':"

    def finalize_and_validate_configuration(self, instance):
        if self.default:
            try:
                instance._get_config(self)
            except KeyError:
                instance._set_config(self.default)

        if self.required and instance._get_config(self) is None:
            name = instance._descriptor_to_name(self)
            prefix = self._error_prefix(instance)
            raise ValueError("{prefix} {name} is a required configuration setting, but no value was set")



from __future__ import annotations

from typing import TYPE_CHECKING

from clearskies.configs import config

if TYPE_CHECKING:
    from clearskies import typing


class Validators(config.Config):
    """
    Validator config.

    A config that accepts various things that are accepted as validators in model columns:

     1. An instance of clearskies.validators.Validator
     2. A list of clearskies.validators.Validator

    Incoming values are normalized to a list so that a list always comes out even if a non-list is provided.
    """

    def __set__(
        self,
        instance,
        value: typing.validator | list[typing.validator],
    ):
        if not isinstance(value, list):
            value = [value]

        for index, item in enumerate(value):
            if hasattr(item, "additional_write_columns") and hasattr(item, "check"):
                continue

            error_prefix = self._error_prefix(instance)
            raise TypeError(
                f"{error_prefix} attempt to set a value of type '{item.__class__.__name__}' for item #{index + 1} when a Validator is required"
            )

        instance._set_config(self, [*value])

    def __get__(self, instance, parent) -> list[typing.validator]:
        if not instance:
            return self  # type: ignore
        return instance._get_config(self)



from typing import Any

from ...schema import Array, Object
from .parameter import Parameter
from .response import Response


class Request:
    formatted_responses: Any = None
    request: Any = None
    relative_path: str = ""

    def __init__(self, oai3_schema_resolver):
        self.oai3_schema_resolver = oai3_schema_resolver

    def set_request(self, request):
        self.request = request
        self.formatted_responses = [self.format_response(response) for response in self.request.responses]
        self.formatted_parameters = [
            self.format_parameter(parameter) for parameter in self.request.parameters if not parameter.in_body
        ]
        self.json_body_parameters = [parameter for parameter in self.request.parameters if parameter.in_body]
        self.relative_path = self.request.relative_path

    def format_response(self, response):
        formatted = Response(self.oai3_schema_resolver)
        formatted.set_response(response)
        return formatted

    def format_parameter(self, parameter):
        formatted = Parameter(self.oai3_schema_resolver)
        formatted.set_parameter(parameter)
        return formatted

    def convert(self):
        data = {}
        for request_method in self.request.request_methods:
            data[request_method.lower()] = {
                "summary": self.request.description,
                "parameters": [parameter.convert() for parameter in self.formatted_parameters],
                "responses": {str(response.status_code): response.convert() for response in self.formatted_responses},
            }

            if self.request.root_properties:
                data[request_method.lower()] = {**data[request_method.lower()], **self.request.root_properties}

            if self.json_body_parameters:
                # For OAI3, there should only be one JSON body root parameter, so it should either be an
                # object or an array.  If we have an array then wrap it in an object
                if type(self.json_body_parameters) == list:
                    definitions = [parameter.definition for parameter in self.json_body_parameters]
                    json_body = Object("body", definitions)
                    is_required = len([1 for param in self.json_body_parameters if param.required]) >= 1
                else:
                    json_body = self.json_body_parameters[0].definition
                    is_required = len([1 for param in json_body.definition.children if param.required]) >= 1

                data[request_method.lower()]["requestBody"] = {
                    "description": self.request.description,
                    "required": is_required,
                    "content": {
                        "application/json": {
                            "schema": self.oai3_schema_resolver(json_body).convert(),
                        },
                    },
                }

        return data



from typing import Any

from clearskies.di.injectable import Injectable


class Di(Injectable):
    def __init__(self):
        pass

    def __get__(self, instance, parent) -> Any:
        if instance is None:
            return self  # type: ignore
        return self._di



import datetime
from types import ModuleType
from typing import Any, Callable
from wsgiref.simple_server import make_server
from wsgiref.util import setup_testing_defaults

import clearskies.endpoint
import clearskies.endpoint_group
from clearskies.contexts.context import Context
from clearskies.di import AdditionalConfig
from clearskies.input_outputs import Wsgi as WsgiInputOutput


class Wsgi(Context):
    def __call__(self, env, start_response):  # type: ignore
        return self.execute_application(WsgiInputOutput(env, start_response))



import datetime

from clearskies.validators.after_column import AfterColumn


class BeforeColumn(AfterColumn):
    def date_comparison(self, incoming_date: datetime.datetime, comparison_date: datetime.datetime, column_name) -> str:
        if incoming_date == comparison_date:
            return "" if self.allow_equal else f"'{column_name}' must be before '{self.other_column_name}'"

        if incoming_date > comparison_date:
            return f"'{column_name}' must be before '{self.other_column_name}'"
        return ""



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Self, overload

import clearskies.parameters_to_properties
import clearskies.typing
from clearskies import configs
from clearskies.autodoc.schema import Array as AutoDocArray
from clearskies.autodoc.schema import Object as AutoDocObject
from clearskies.autodoc.schema import Schema as AutoDocSchema
from clearskies.column import Column
from clearskies.di.inject import InputOutput
from clearskies.functional import string, validations

if TYPE_CHECKING:
    from clearskies import Column, Model


class HasMany(Column):
    """
    A column to manage a "has many" relationship.

    In order to manage a has-many relationship, the child model needs a column that stores the
    id of the parent record it belongs to.  Also remember that the reverse of a has-many relationship
    is a belongs-to relationship: the parent has many children, the child belongs to a parent.

    There's an automatic standard where the name of the column in thie child table that stores the
    parent id is made by converting the parent model class name into snake case and then appending
    `_id`.  For instance, if the parent model is called the `DooHicky` class, the child model is
    expected to have a column named `doo_hicky_id`.  If you use a different column name for the
    id in your child model, then just update the `foreign_column_name` property on the `HasMany`
    column accordingly.

    See the BelongsToId class for additional background and directions on avoiding circular dependency trees.

    ```python
    import clearskies


    class Product(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        category_id = clearskies.columns.String()


    class Category(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        products = clearskies.columns.HasMany(Product)


    def test_has_many(products: Product, categories: Category):
        toys = categories.create({"name": "Toys"})
        auto = categories.create({"name": "Auto"})

        # create some toys
        ball = products.create({"name": "Ball", "category_id": toys.id})
        fidget_spinner = products.create({"name": "Fidget Spinner", "category_id": toys.id})
        crayon = products.create({"name": "Crayon", "category_id": toys.id})

        # the HasMany column is an interable of matching records
        toy_names = [product.name for product in toys.products]

        # it specifically returns a models object so you can do more filtering/transformations
        return toys.products.sort_by("name", "asc")


    cli = clearskies.contexts.Cli(
        clearskies.endpoints.Callable(
            test_has_many,
            model_class=Product,
            readable_column_names=["id", "name"],
        ),
        classes=[Category, Product],
    )

    if __name__ == "__main__":
        cli()
    ```

    And if you execute this it will return:

    ```json
    {
        "status": "success",
        "error": "",
        "data": [
            {"id": "edc68e8d-7fc8-45ce-98f0-9c6f883e4e7f", "name": "Ball"},
            {"id": "b51a0de5-c784-4e0c-880c-56e5bf731dfd", "name": "Crayon"},
            {"id": "06cec3af-d042-4d6b-a99c-b4a0072f188d", "name": "Fidget Spinner"},
        ],
        "pagination": {},
        "input_errors": {},
    }
    ```
    """

    """
    HasMany columns are not currently writeable.
    """
    is_writeable = configs.Boolean(default=False)
    is_searchable = configs.Boolean(default=False)
    _descriptor_config_map = None

    """ The model class for the child table we keep our "many" records in. """
    child_model_class = configs.ModelClass(required=True)

    """
    The name of the column in the child table that connects it back to the parent.

    By default this is populated by converting the model class name from TitleCase to snake_case and appending _id.
    So, if the model class is called `ProductCategory`, this becomes `product_category_id`.  This MUST correspond to
    the actual name of a column in the child table.  This is used so that the parent can find its child records.

    Example:

    ```python
    import clearskies

    class Product(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        my_parent_category_id = clearskies.columns.String()

    class Category(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        products = clearskies.columns.HasMany(Product, foreign_column_name="my_parent_category_id")

    def test_has_many(products: Product, categories: Category):
        toys = categories.create({"name": "Toys"})

        fidget_spinner = products.create({"name": "Fidget Spinner", "my_parent_category_id": toys.id})
        crayon = products.create({"name": "Crayon", "my_parent_category_id": toys.id})
        ball = products.create({"name": "Ball", "my_parent_category_id": toys.id})

        return toys.products.sort_by("name", "asc")

    cli = clearskies.contexts.Cli(
        clearskies.endpoints.Callable(
            test_has_many,
            model_class=Product,
            readable_column_names=["id", "name"],
        ),
        classes=[Category, Product],
    )

    if __name__ == "__main__":
        cli()
    ```

    Compare to the first example for the HasMany class.  In that case, the column in the product model which
    contained the category id was `category_id`, and the `products` column didn't have to specify the
    `foreign_column_name` (since the column name followed the naming rule).  As a result, `category.products`
    was able to find all children of a given category.  In this example, the name of the column in the product
    model that contains the category id was changed to `my_parent_category_id`.  Since this no longer matches
    the naming convention, we had to specify `foreign_column_name="my_parent_category_id"` in `Category.products`,
    in order for the `HasMany` column to find the children.  Therefore, when invoked it returns the same thing:

    ```json
    {
        "status": "success",
        "error": "",
        "data": [
            {
                "id": "3cdd06e0-b226-4a4a-962d-e8c5acc759ac",
                "name": "Ball"
            },
            {
                "id": "debc7968-976a-49cd-902c-d359a8abd032",
                "name": "Crayon"
            },
            {
                "id": "0afcd314-cdfc-4a27-ac6e-061b74ee5bf9",
                "name": "Fidget Spinner"
            }
        ],
        "pagination": {},
        "input_errors": {}
    }
    ```
    """
    foreign_column_name = configs.ModelToIdColumn()

    """
    Columns from the child table that should be included when converting this column to JSON.

    You can tell an endpoint to include a `HasMany` column in the response.  If you do this, the columns
    from the child class that are included in the JSON response are determined by `readable_child_column_names`.
    Example:

    ```python
    import clearskies

    class Product(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        category_id = clearskies.columns.String()

    class Category(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        products = clearskies.columns.HasMany(Product, readable_child_column_names=["id", "name"])

    def test_has_many(products: Product, categories: Category):
        toys = categories.create({"name": "Toys"})

        fidget_spinner = products.create({"name": "Fidget Spinner", "category_id": toys.id})
        ball = products.create({"name": "Ball", "category_id": toys.id})
        crayon = products.create({"name": "Crayon", "category_id": toys.id})

        return toys

    cli = clearskies.contexts.Cli(
        clearskies.endpoints.Callable(
            test_has_many,
            model_class=Category,
            readable_column_names=["id", "name", "products"],
        ),
        classes=[Category, Product],
    )

    if __name__ == "__main__":
        cli()
    ```

    In this example we're no longer returning a list of products directly.  Instead, we're returning a query
    on the categories nodel and asking the endpoint to also unpack their products.  We set `readable_child_column_names`
    to `["id", "name"]` for `Category.products`, so when the endpoint unpacks the products, it includes those columns:

    ```json
    {
        "status": "success",
        "error": "",
        "data": [
            {
                "id": "c8a71c81-fa0e-427d-a166-159f3c9de72b",
                "name": "Office Supplies",
                "products": [
                    {
                        "id": "6d24ffa2-6e1b-4ce9-87ff-daf2ba237c92",
                        "name": "Stapler"
                    },
                    {
                        "id": "3a42cd7d-6804-465e-9fb1-055fafa7fc62",
                        "name": "Chair"
                    }
                ]
            },
            {
                "id": "5a790950-858b-411a-bf5c-1338a28e73d0",
                "name": "Toys",
                "products": [
                    {
                        "id": "d4022224-cc22-49c2-8da9-7a8f9fa7e976",
                        "name": "Fidget Spinner"
                    },
                    {
                        "id": "415fa48e-984a-4703-b6e6-f88f741403c8",
                        "name": "Ball"
                    },
                    {
                        "id": "58328363-5180-441c-b1a8-1b92e12a8f08",
                        "name": "Crayon"
                    }
                ]
            }
        ],
        "pagination": {},
        "input_errors": {}
    }

    ```

    """
    readable_child_column_names = configs.ReadableModelColumns("child_model_class")

    """
    Additional conditions to add to searches on the child table.

    ```python
    import clearskies

    class Order(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        total = clearskies.columns.Float()
        status = clearskies.columns.Select(["Open", "In Progress", "Closed"])
        user_id = clearskies.columns.String()

    class User(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        orders = clearskies.columns.HasMany(Order, readable_child_column_names=["id", "status"])
        large_open_orders = clearskies.columns.HasMany(
            Order,
            readable_child_column_names=["id", "status"],
            where=[Order.status.equals("Open"), "total>100"],
        )

    def test_has_many(users: User, orders: Order):
        user = users.create({"name": "Bob"})

        order_1 = orders.create({"status": "Open", "total": 25.50, "user_id": user.id})
        order_2 = orders.create({"status": "Closed", "total": 35.50, "user_id": user.id})
        order_3 = orders.create({"status": "Open", "total": 125, "user_id": user.id})
        order_4 = orders.create({"status": "In Progress", "total": 25.50, "user_id": user.id})

        return user.large_open_orders

    cli = clearskies.contexts.Cli(
        clearskies.endpoints.Callable(
            test_has_many,
            model_class=Order,
            readable_column_names=["id", "total", "status"],
            return_records=True,
        ),
        classes=[Order, User],
    )

    if __name__ == "__main__":
        cli()
    ```

    The above example shows two different ways of adding conditions.  Note that `where` can be either a list or a single
    condition.  If you invoked this you would get:

    ```json
    {
        "status": "success",
        "error": "",
        "data": [
            {
                "id": "6ad99935-ac9a-40ef-a1b2-f34538cc6529",
                "total": 125.0,
                "status": "Open"
            }
        ],
        "pagination": {},
        "input_errors": {}
    }
    ```

    Finally, an individual condition can also be a callable that accepts the child model class, adds any desired conditions,
    and then returns the modified model class.  Like usual, this callable can request any defined depenency.  So, for
    instance, the following column definition is equivalent to the example above:

    ```python
    class User(clearskies.Model):
        # removing unchanged part for brevity
        large_open_orders = clearskies.columns.HasMany(
            Order,
            readable_child_column_names=["id", "status"],
            where=lambda model: model.where("status=Open").where("total>100"),
        )
    ```
    """
    where = configs.Conditions()

    input_output = InputOutput()

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        child_model_class,
        foreign_column_name: str | None = None,
        readable_child_column_names: list[str] = [],
        where: clearskies.typing.condition | list[clearskies.typing.condition] = [],
        is_readable: bool = True,
        on_change_pre_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_post_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_save_finished: clearskies.typing.action | list[clearskies.typing.action] = [],
    ):
        pass

    def finalize_configuration(self, model_class, name) -> None:
        """
        Finalize and check the configuration.

        This is an external trigger called by the model class when the model class is ready.
        The reason it exists here instead of in the constructor is because some columns are tightly
        connected to the model class, and can't validate configuration until they know what the model is.
        Therefore, we need the model involved, and the only way for a property to know what class it is
        in is if the parent class checks in (which is what happens here).
        """
        # this is where we auto-calculate the expected name of our id column in the child model.
        # we can't do it until now because it comes from the model class we are connected to, and
        # we only just get it.
        foreign_column_name_config = self._get_config_object("foreign_column_name")
        foreign_column_name_config.set_model_class(self.child_model_class)
        has_value = False
        try:
            has_value = bool(self.foreign_column_name)
        except KeyError:
            pass

        if not has_value:
            self.foreign_column_name = string.camel_case_to_snake_case(model_class.__name__) + "_id"

        super().finalize_configuration(model_class, name)

    @property
    def child_columns(self) -> dict[str, Column]:
        return self.child_model_class.get_columns()

    @property
    def child_model(self) -> Model:
        return self.di.build(self.child_model_class, cache=True)

    @overload
    def __get__(self, instance: None, cls: type[Model]) -> Self:
        pass

    @overload
    def __get__(self, instance: Model, cls: type[Model]) -> Model:
        pass

    def __get__(self, model, cls):
        if model is None:
            self.model_class = cls
            return self  # type:  ignore

        # this makes sure we're initialized
        if "name" not in self._config:  # type: ignore
            model.get_columns()

        foreign_column_name = self.foreign_column_name
        model_id = getattr(model, model.id_column_name)
        children = self.child_model.where(f"{foreign_column_name}={model_id}")

        if not self.where:
            return children

        for index, where in enumerate(self.where):
            if callable(where):
                children = self.di.call_function(where, model=children, **self.input_output.get_context_for_callables())
                if not validations.is_model(children):
                    raise ValueError(
                        f"Configuration error for column '{self.name}' in model '{self.model_class.__name__}': when 'where' is a callable, it must return a models class, but when the callable in where entry #{index + 1} was called, it did not return the models class"
                    )
            else:
                children = children.where(where)
        return children

    def __set__(self, model: Model, value: Model) -> None:
        raise ValueError(
            f"Attempt to set a value to {model.__class__.__name__}.{self.name}: this is not allowed because it is a HasMany column, which is not writeable."
        )

    def to_json(self, model: Model) -> dict[str, Any]:
        children = []
        columns = self.child_columns
        child_id_column_name = self.child_model_class.id_column_name
        json: dict[str, Any] = {}
        for child in getattr(model, self.name):
            json = {
                **json,
                **columns[child_id_column_name].to_json(child),
            }
            for column_name in self.readable_child_column_names:
                json = {
                    **json,
                    **columns[column_name].to_json(child),
                }
            children.append(json)
        return {self.name: children}

    def documentation(
        self, name: str | None = None, example: str | None = None, value: str | None = None
    ) -> list[AutoDocSchema]:
        columns = self.child_columns
        child_id_column_name = self.child_model.id_column_name
        child_properties = [columns[child_id_column_name].documentation()]

        for column_name in self.readable_child_column_names:
            child_properties.extend(columns[column_name].documentation())  # type: ignore

        child_object = AutoDocObject(
            string.title_case_to_nice(self.child_model_class.__name__),
            child_properties,
        )
        return [AutoDocArray(name if name is not None else self.name, child_object, value=value)]



from __future__ import annotations

from typing import TYPE_CHECKING

from clearskies.configs import config

if TYPE_CHECKING:
    from clearskies import typing


class Joins(config.Config):
    def __set__(self, instance, value: typing.join | list[typing.join]):
        if not isinstance(value, list):
            value = [value]

        for index, item in enumerate(value):
            if callable(item) or isinstance(item, str):
                continue

            error_prefix = self._error_prefix(instance)
            raise TypeError(
                f"{error_prefix} attempt to set a value of type '{item.__class__.__name__}' for item #{index + 1} when a string or callable is required"  # type: ignore
            )

        instance._set_config(self, [*value])

    def __get__(self, instance, parent) -> list[typing.join]:
        if not instance:
            return self  # type: ignore
        return instance._get_config(self)



import re


class Headers:
    _headers: dict[str, str] = {}

    def __init__(self, headers: dict[str, str] = {}):
        self.__dict__["_headers"] = (
            {key.upper().replace("_", "-"): value for (key, value) in headers.items()} if headers else {}
        )

    def __contains__(self, key: str):
        return key.upper().replace("_", "-") in self._headers

    def __getattr__(self, key: str) -> str | None:
        return self._headers.get(key.upper().replace("_", "-"), None)

    def __setattr__(self, key: str, value: str) -> None:
        if not isinstance(key, str):
            raise TypeError(
                f"Header keys must be strings, but an object of type '{value.__class__.__name__}' was provided."
            )
        if not isinstance(value, str):
            raise TypeError(
                f"Header values must be strings, but an object of type '{value.__class__.__name__}' was provided."
            )
        self._headers[re.sub("\\s+", " ", key.upper().replace("_", "-"))] = re.sub("\\s+", " ", value.strip())

    def get(self, key, default=None):
        if key not in self:
            return default
        return self.__getattr__(key)

    def keys(self) -> list[str]:
        return list(self._headers.keys())

    def values(self) -> list[str]:
        return list(self._headers.keys())

    def items(self) -> list[tuple[str]]:
        return list(self._headers.items())  # type: ignore

    def add(self, key: str, value: str) -> None:
        """Add a header.  This expects a string with a colon separating the key and value."""
        setattr(self, key, value)



import clearskies.configs
import clearskies.parameters_to_properties
from clearskies.security_header import SecurityHeader


class Cors(SecurityHeader):
    origin = clearskies.configs.String()
    methods = clearskies.configs.StringList(default=[])
    headers = clearskies.configs.StringList(default=[])
    max_age = clearskies.configs.Integer(default=5)
    credentials = clearskies.configs.Boolean(default=False)
    expose_headers = clearskies.configs.StringList(default=[])
    is_cors = True

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        credentials: bool = False,
        expose_headers: list[str] = [],
        headers: list[str] = [],
        max_age: int = 5,
        methods: list[str] = [],
        origin: str = "",
    ):
        self.finalize_and_validate_configuration()

    def set_headers(self, headers: list[str]):
        self.headers = headers

    def add_header(self, header: str):
        self.headers = [*self.headers, header]

    def set_methods(self, methods: list[str]):
        self.methods = methods

    def add_method(self, method: str):
        self.methods = [*self.methods, method]

    def set_headers_for_input_output(self, input_output):
        for key in ["expose_headers", "methods", "headers"]:
            value = getattr(self, key)
            if not value:
                continue
            input_output.response_headers.add(f"access-control-allow-{key}".replace("_", "-"), ", ".join(value))
        if self.credentials:
            input_output.response_headers.add("access-control-allow-credentials", "true")
        if self.max_age:
            input_output.response_headers.add("access-control-max-age", str(self.max_age))
        if self.origin:
            input_output.response_headers.add("access-control-allow-origin", str(self.origin))



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable

import clearskies.parameters_to_properties
import clearskies.typing
from clearskies import configs
from clearskies.columns.belongs_to_id import BelongsToId

if TYPE_CHECKING:
    from clearskies import Model


class CategoryTree(BelongsToId):
    """
    The category tree helps you do quick lookups on a typical category tree.

    It's a very niche tool.  In general, graph databases solve this problem better, but
    it's not always worth the effort of spinning up a new kind of database.

    This column needs a special tree table where it will pre-compute and store the
    necessary information to perform quick lookups about relationships in a cateogry
    tree.  So, imagine you have a table that represents a standard category heirarchy:

    ```sql
    CREATE TABLE categories (
      id varchar(255),
      parent_id varchar(255),
      name varchar(255)
    )

    `parent_id`, in this case, would be a reference to the `categories` table itself -
    hence the heirarchy.  This works fine  as a starting point but it gets tricky when you want to answer questions like
    "what are all the parent categories of category X?" or "what are all the child categories of category Y?".
    This column class solves that by building a tree table that caches this data as the categories are updated.
    That table should look like this:

    ```sql
    CREATE TABLE category_tree (
      id varchar(255),
      parent_id varchar(255),
      child_id varchar(255),
      is_parent tinyint(1),
      level tinyint(1),
    )
    ```

    Then you would attach this column to your category model as a replacement for a typical BelongsToId relationship:

    ```python
    import clearskies

    class Tree(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend(silent_on_missing_tables=True)

        id = clearskies.columns.Uuid()
        parent_id = clearskies.columns.String()
        child_id = clearskies.columns.String()
        is_parent = clearskies.columns.Boolean()
        level = clearskies.columns.Integer()

    class Category(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend(silent_on_missing_tables=True)

        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        parent_id = clearskies.columns.CategoryTree(Tree)
        parent = clearskies.columns.BelongsToModel("parent_id")
        children = clearskies.columns.CategoryTreeChildren("parent_id")
        descendants = clearskies.columns.CategoryTreeDescendants("parent_id")
        ancestors = clearskies.columns.CategoryTreeAncestors("parent_id")

    def test_category_tree(category: Category):
        root_1 = category.create({"name": "Root 1"})
        root_2 = category.create({"name": "Root 2"})
        sub_1_root_1 = category.create({"name": "Sub 1 of Root 1", "parent_id": root_1.id})
        sub_2_root_1 = category.create({"name": "Sub 2 of Root 1", "parent_id": root_1.id})
        sub_sub = category.create({"name": "Sub Sub", "parent_id": sub_1_root_1.id})
        sub_1_root_2 = category.create({"name": "Sub 1 of Root 2", "parent_id": root_2.id})

        return {
            "descendants_of_root_1": [descendant.name for descendant in root_1.descendants],
            "children_of_root_1": [child.name for child in root_1.children],
            "descendants_of_root_2": [descendant.name for descendant in root_2.descendants],
            "ancestors_of_sub_sub": [ancestor.name for ancestor in sub_sub.ancestors],
        }

    cli = clearskies.contexts.Cli(
        clearskies.endpoints.Callable(test_category_tree),
        classes=[Category, Tree],
    )
    cli()
    ```

    And if you invoke the above you will get:

    ```json
    {
        "status": "success",
        "error": "",
        "data": {
            "descendants_of_root_1": ["Sub 1 of Root 1", "Sub 2 of Root 1", "Sub Sub"],
            "children_of_root_1": ["Sub 1 of Root 1", "Sub 2 of Root 1"],
            "descendants_of_root_2": ["Sub 1 of Root 2"],
            "ancestors_of_sub_sub": ["Root 1", "Sub 1 of Root 1"],
        },
        "pagination": {},
        "input_errors": {},
    }
    ```

    In case it's not clear, the definition of these things are:

     1. Descendants: All children under a given category (recursively).
     2. Children: The direct descendants of a given category.
     3. Ancestors: The parents of a given category, starting from the root category.
     4. Parent: the immediate parent of the category.

    """

    """
    The model class that will persist our tree data
    """
    tree_model_class = configs.ModelClass(required=True)

    """
    The column in the tree model that references the parent in the relationship
    """
    tree_parent_id_column_name = configs.ModelColumn("tree_model_class", default="parent_id")

    """
    The column in the tree model that references the child in the relationship
    """
    tree_child_id_column_name = configs.ModelColumn("tree_model_class", default="child_id")

    """
    The column in the tree model that denotes which node in the relationship represents the tree
    """
    tree_is_parent_column_name = configs.ModelColumn("tree_model_class", default="is_parent")

    """
    The column in the tree model that references the parent in a relationship
    """
    tree_level_column_name = configs.ModelColumn("tree_model_class", default="level")

    """
    The maximum expected depth of the tree
    """
    max_iterations = configs.Integer(default=100)

    """
    The strategy for loading relatives.

    Choose whatever one actually works for your backend

     * JOIN: use an actual `JOIN` (e.g. quick and efficient, but mostly only works for SQL backends).
     * WHERE IN: Use a `WHERE IN` condition.
     * INDIVIDUAL: Load each record separately.  Works for any backend but is also the slowest.
    """
    load_relatives_strategy = configs.Select(["join", "where_in", "individual"], default="join")

    _descriptor_config_map = None

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        tree_model_class,
        tree_parent_id_column_name: str = "parent_id",
        tree_child_id_column_name: str = "child_id",
        tree_is_parent_column_name: str = "is_parent",
        tree_level_column_name: str = "level",
        max_iterations: int = 100,
        load_relatives_strategy: str = "join",
        readable_parent_columns: list[str] = [],
        join_type: str | None = None,
        where: clearskies.typing.condition | list[clearskies.typing.condition] = [],
        default: str | None = None,
        setable: str | Callable | None = None,
        is_readable: bool = True,
        is_writeable: bool = True,
        is_searchable: bool = True,
        is_temporary: bool = False,
        validators: clearskies.typing.validator | list[clearskies.typing.validator] = [],
        on_change_pre_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_post_save: clearskies.typing.action | list[clearskies.typing.action] = [],
        on_change_save_finished: clearskies.typing.action | list[clearskies.typing.action] = [],
        created_by_source_type: str = "",
        created_by_source_key: str = "",
        created_by_source_strict: bool = True,
    ):
        pass

    def finalize_configuration(self, model_class, name) -> None:
        """
        Finalize and check the configuration.

        This is an external trigger called by the model class when the model class is ready.
        The reason it exists here instead of in the constructor is because some columns are tightly
        connected to the model class, and can't validate configuration until they know what the model is.
        Therefore, we need the model involved, and the only way for a property to know what class it is
        in is if the parent class checks in (which is what happens here).
        """
        self.parent_model_class = model_class
        super().finalize_configuration(model_class, name)

    @property
    def tree_model(self):
        return self.di.build(self.tree_model_class, cache=True)

    def post_save(self, data: dict[str, Any], model: Model, id: int | str) -> None:
        if not model.is_changing(self.name, data):
            return

        self.update_tree_table(model, id, model.latest(self.name, data))
        return

    def force_tree_update(self, model: Model):
        self.update_tree_table(model, getattr(model, model.id_column_name), getattr(model, self.name))

    def update_tree_table(self, model: Model, child_id: int | str, direct_parent_id: int | str) -> None:
        tree_model = self.tree_model
        parent_model = self.parent_model
        tree_parent_id_column_name = self.tree_parent_id_column_name
        tree_child_id_column_name = self.tree_child_id_column_name
        tree_is_parent_column_name = self.tree_is_parent_column_name
        tree_level_column_name = self.tree_level_column_name
        max_iterations = self.max_iterations

        # we're going to be lazy and just delete the data for the current record in the tree table,
        # and then re-insert everything (but we can skip this if creating a new record)
        if model:
            for tree in tree_model.where(f"{tree_child_id_column_name}={child_id}"):
                tree.delete()

        # if we are a root category then we don't have a tree
        if not direct_parent_id:
            return

        is_root = False
        id_column_name = parent_model.id_column_name
        next_parent = parent_model.find(f"{id_column_name}={direct_parent_id}")
        tree = []
        c = 0
        while not is_root:
            c += 1
            if c > max_iterations:
                self._circular(max_iterations)

            tree.append(getattr(next_parent, next_parent.id_column_name))
            if not getattr(next_parent, self.name):
                is_root = True
            else:
                next_next_parent_id = getattr(next_parent, self.name)
                next_parent = model.find(f"{id_column_name}={next_next_parent_id}")

        tree.reverse()
        for index, parent_id in enumerate(tree):
            tree_model.create(
                {
                    tree_parent_id_column_name: parent_id,
                    tree_child_id_column_name: child_id,
                    tree_is_parent_column_name: 1 if parent_id == direct_parent_id else 0,
                    tree_level_column_name: index,
                }
            )

    def _circular(self, max_iterations):
        raise ValueError(
            f"Error for column {self.model_class.__name__}.{self.name}: "
            + f"I've climbed through {max_iterations} parents and haven't found the root yet."
            + "You may have accidentally created a circular cateogry tree.  If not, and your category tree "
            + "really _is_ that deep, then adjust the 'max_iterations' configuration for this column accordingly. "
        )



from clearskies.exceptions.authentication import Authentication
from clearskies.exceptions.authorization import Authorization
from clearskies.exceptions.client_error import ClientError
from clearskies.exceptions.input_errors import InputErrors
from clearskies.exceptions.moved_permanently import MovedPermanently
from clearskies.exceptions.moved_temporarily import MovedTemporarily
from clearskies.exceptions.not_found import NotFound

__all__ = [
    "Authentication",
    "Authorization",
    "ClientError",
    "InputErrors",
    "MovedPermanently",
    "MovedTemporarily",
    "NotFound",
]



from __future__ import annotations

import inspect
from collections import OrderedDict
from typing import TYPE_CHECKING, Any, Callable, Type

import clearskies.configs
import clearskies.exceptions
from clearskies import authentication, autodoc, typing
from clearskies.endpoints.get import Get
from clearskies.functional import routing, string
from clearskies.input_outputs import InputOutput

if TYPE_CHECKING:
    from clearskies import SecurityHeader
    from clearskies.model import Column, Model, Schema


class Update(Get):
    """
    An endpoint to update a record.

    This endpoint handles update operations.  As with the `Get` endpoint, it will lookup the record by taking
    the record id (or any other unique column you specify) out of the URL and then will fetch that record
    using the model class.  Then, it will use the model and list of writeable column names to validate the
    incoming user input.  The default request method is `PATCH`.  If everything checks out, it will then
    update the record.

    ```python
    import clearskies


    class User(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()
        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()
        username = clearskies.columns.String(validators=[clearskies.validators.Required()])


    wsgi = clearskies.contexts.WsgiRef(
        clearskies.endpoints.Update(
            model_class=User,
            url="/{id}",
            readable_column_names=["id", "name", "username"],
            writeable_column_names=["name", "username"],
        ),
        bindings={
            "memory_backend_default_data": [
                {
                    "model_class": User,
                    "records": [
                        {"id": "1-2-3-4", "name": "Bob Brown", "username": "bobbrown"},
                        {"id": "1-2-3-5", "name": "Jane Doe", "username": "janedoe"},
                        {"id": "1-2-3-6", "name": "Greg", "username": "greg"},
                    ],
                },
            ],
        },
    )
    wsgi()
    ```

    And when invoked:

    ```bash
    $ curl 'http://localhost:8080/1-2-3-4' -X PATCH -d '{"name": "Bobby Brown", "username": "bobbybrown"}' | jq
    {
        "status": "success",
        "error": "",
        "data": {
            "id": "1-2-3-4",
            "name": "Bobby Brown",
            "username": "bobbybrown"
        },
        "pagination": {},
        "input_errors": {}
    }

    $ curl 'http://localhost:8080/1-2-3-5' -X PATCH -d '{"name": 12345, "username": ""}' | jq
    {
        "status": "input_errors",
        "error": "",
        "data": [],
        "pagination": {},
        "input_errors": {
            "name": "value should be a string",
            "username": "'username' is required."
        }
    }
    """

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        model_class: type[Model],
        url: str,
        writeable_column_names: list[str],
        readable_column_names: list[str],
        record_lookup_column_name: str | None = None,
        input_validation_callable: Callable | None = None,
        request_methods: list[str] = ["PATCH"],
        response_headers: list[str | Callable[..., list[str]]] = [],
        output_map: Callable[..., dict[str, Any]] | None = None,
        output_schema: Schema | None = None,
        column_overrides: dict[str, Column] = {},
        internal_casing: str = "snake_case",
        external_casing: str = "snake_case",
        security_headers: list[SecurityHeader] = [],
        description: str = "",
        where: typing.condition | list[typing.condition] = [],
        joins: typing.join | list[typing.join] = [],
        authentication: authentication.Authentication = authentication.Public(),
        authorization: authentication.Authorization = authentication.Authorization(),
    ):
        # see comment in clearskies.endpoints.Create.__init__
        self.request_methods = request_methods

        # we need to call the parent but don't have to pass along any of our kwargs.  They are all optional in our parent, and our parent class
        # just stores them in parameters, which we have already done.  However, the parent does do some extra initialization stuff that we need,
        # which is why we have to call the parent.
        super().__init__(model_class, url, readable_column_names)

    def handle(self, input_output: InputOutput) -> Any:
        request_data = self.get_request_data(input_output)
        if not request_data and input_output.has_body():
            raise clearskies.exceptions.ClientError("Request body was not valid JSON")
        model = self.fetch_model(input_output)
        self.validate_input_against_schema(request_data, input_output, model)
        model.save(request_data)
        return self.success(input_output, self.model_as_json(model, input_output))

    def documentation(self) -> list[autodoc.request.Request]:
        output_schema = self.model_class
        nice_model = string.camel_case_to_words(output_schema.__name__)

        schema_model_name = string.camel_case_to_snake_case(output_schema.__name__)
        output_data_schema = self.documentation_data_schema(output_schema, self.readable_column_names)
        output_autodoc = (
            autodoc.schema.Object(
                self.auto_case_internal_column_name("data"), children=output_data_schema, model_name=schema_model_name
            ),
        )

        authentication = self.authentication
        standard_error_responses = [self.documentation_input_error_response()]
        if not getattr(authentication, "is_public", False):
            standard_error_responses.append(self.documentation_access_denied_response())
            if getattr(authentication, "can_authorize", False):
                standard_error_responses.append(self.documentation_unauthorized_response())

        return [
            autodoc.request.Request(
                self.description,
                [
                    self.documentation_success_response(
                        output_autodoc,  # type: ignore
                        description=self.description,
                    ),
                    *standard_error_responses,
                    self.documentation_generic_error_response(),
                ],
                relative_path=self.url,
                request_methods=self.request_methods,
                parameters=[
                    *self.documentation_request_parameters(),
                    *self.standard_url_parameters(),
                ],
                root_properties={
                    "security": self.documentation_request_security(),
                },
            ),
        ]

    def documentation_request_parameters(self) -> list[autodoc.request.Parameter]:
        return [
            *self.standard_json_request_parameters(self.model_class),
            *self.standard_url_request_parameters(),
        ]

    def documentation_models(self) -> dict[str, autodoc.schema.Schema]:
        output_schema = self.output_schema if self.output_schema else self.model_class
        schema_model_name = string.camel_case_to_snake_case(output_schema.__name__)

        return {
            schema_model_name: autodoc.schema.Object(
                self.auto_case_internal_column_name("data"),
                children=self.documentation_data_schema(output_schema, self.readable_column_names),
            ),
        }



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Any

import clearskies.configs
from clearskies.validator import Validator

if TYPE_CHECKING:
    import clearskies.model


class MinimumValue(Validator):
    minimum_value = clearskies.configs.Integer(required=True)

    def __init__(self, minimum_value: int):
        self.minimum_value = minimum_value
        self.finalize_and_validate_configuration()

    def check(self, model: clearskies.model.Model, column_name: str, data: dict[str, Any]) -> str:
        if column_name not in data:
            return ""
        try:
            value = float(data[column_name])
        except ValueError:
            return f"{column_name} must be an integer or float"
        if float(value) >= self.minimum_value:
            return ""
        return f"'{column_name}' must be at least {self.minimum_value}."



from __future__ import annotations

from collections import OrderedDict
from typing import TYPE_CHECKING, Any, Self, overload

import clearskies.parameters_to_properties
from clearskies import configs
from clearskies.column import Column
from clearskies.columns.belongs_to_id import BelongsToId
from clearskies.functional import validations

if TYPE_CHECKING:
    from clearskies import Model


class BelongsToModel(Column):
    """Return the model object for a belongs to relationship."""

    """ The name of the belongs to column we are connected to. """
    belongs_to_column_name = configs.ModelColumn(required=True)

    is_temporary = clearskies.configs.boolean.Boolean(default=True)
    _descriptor_config_map = None

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        belongs_to_column_name: str,
    ):
        pass

    def finalize_configuration(self, model_class: type, name: str) -> None:
        """Finalize and check the configuration."""
        getattr(self.__class__, "belongs_to_column_name").set_model_class(model_class)
        self.model_class = model_class
        self.name = name
        self.finalize_and_validate_configuration()

        # finally, let the belongs to column know about us and make sure it's the right thing.
        belongs_to_column = getattr(model_class, self.belongs_to_column_name)
        if not isinstance(belongs_to_column, BelongsToId):
            raise ValueError(
                f"Error with configuration for {model_class.__name__}.{name}, which is a BelongsToModel.  It needs to point to a belongs to column, and it was told to use {model_class.__name__}.{self.belongs_to_column_name}, but this is not a BelongsToId column."
            )
        belongs_to_column.model_column_name = name

    @overload
    def __get__(self, instance: None, cls: type[Model]) -> Self:
        pass

    @overload
    def __get__(self, instance: Model, cls: type[Model]) -> Model:
        pass

    def __get__(self, model, cls):
        if model is None:
            self.model_class = cls
            return self  # type:  ignore

        # this makes sure we're initialized
        if "name" not in self._config:  # type: ignore
            model.get_columns()

        belongs_to_column = getattr(model.__class__, self.belongs_to_column_name)
        parent_id = getattr(model, self.belongs_to_column_name)
        parent_class = belongs_to_column.parent_model_class
        parent_model = self.di.build(parent_class, cache=False)
        if not parent_id:
            return parent_model.empty_model()

        parent_id_column_name = parent_model.id_column_name
        join_alias = belongs_to_column.join_table_alias()
        raw_data = model.get_raw_data()

        # sometimes the model is loaded via the N+1 functionality, in which case the data will already exist
        # in model.data but hiding under a different name.
        if raw_data.get(f"{join_alias}.{parent_id_column_name}"):
            parent_data = {parent_id_column_name: raw_data[f"{join_alias}_{parent_id_column_name}"]}
            for column_name in belongs_to_column.readable_parent_columns:
                select_alias = f"{join_alias}_{column_name}"
                parent_data[column_name] = raw_data[select_alias] if select_alias in raw_data else None
            return parent_model.model(parent_data)

        return parent_model.find(f"{parent_id_column_name}={parent_id}")

    def __set__(self, model: Model, value: Model) -> None:
        setattr(model, self.belongs_to_column_name, getattr(value, value.id_column_name))

    def pre_save(self, data: dict[str, Any], model: Model) -> dict[str, Any]:
        # if we have a model coming in then we want to extract the id.  Either way, the model id needs to go to the
        # belongs_to_id column, which is the only one that is actually saved.
        if self.name in data:
            value = data[self.name]
            data[self.belongs_to_column_name] = (
                getattr(value, value.id_column_name) if validations.is_model(value) else value
            )
        return super().pre_save(data, model)

    def add_join(self, model: Model) -> Model:
        return getattr(model.__class__, self.belongs_to_column_name).add_join(model)

    def join_table_alias(self) -> str:
        return getattr(self.model_class, self.belongs_to_column_name).join_table_alias()

    def add_search(
        self, model: clearskies.model.Model, value: str, operator: str = "", relationship_reference: str = ""
    ) -> clearskies.model.Model:
        return getattr(self.model_class, self.belongs_to_column_name).add_search(
            model, value, operator, relationship_reference=relationship_reference
        )

    def to_json(self, model: Model) -> dict[str, Any]:
        """Convert the column into a json-friendly representation."""
        belongs_to_column = getattr(model.__class__, self.belongs_to_column_name)
        if not belongs_to_column.readable_parent_columns:
            raise ValueError(
                f"Configuration error for {model.__class__.__name__}: I can't convert to JSON unless you set readable_parent_columns on my parent attribute, {model.__class__.__name__}.{self.belongs_to_column_name}."
            )

        # otherwise return an object with the readable parent columns
        columns = belongs_to_column.parent_columns
        parent = getattr(model, self.name)
        json: dict[str, Any] = OrderedDict()
        for column_name in belongs_to_column.readable_parent_columns:
            json = {**json, **columns[column_name].to_json(parent)}  # type: ignore
        return {
            self.name: json,
        }



import inspect

import wrapt  # type: ignore


@wrapt.decorator
def parameters_to_properties(wrapped, instance, args, kwargs):
    if not instance:
        raise ValueError(
            "The parameters_to_properties decorator only works for methods in classes, not plain functions"
        )

    if args:
        wrapped_args = inspect.getfullargspec(wrapped)
        for key, value in zip(wrapped_args.args[1:], args):
            # if it's a dictionary or a list then copy it to avoid linking data
            if isinstance(value, dict):
                value = {**value}
            if isinstance(value, list):
                value = [*value]
            setattr(instance, key, value)

    for key, value in kwargs.items():
        # if it's a dictionary or a list then copy it to avoid linking data
        if isinstance(value, dict):
            value = {**value}
        if isinstance(value, list):
            value = [*value]
        setattr(instance, key, value)

    wrapped(*args, **kwargs)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Self

from clearskies.di.injectable import Injectable

if TYPE_CHECKING:
    from clearskies.di import Di


class InjectableProperties:
    """
    Fetch dependencies via properties rather than constructor arguments.

    This class allows you to specify dependencies by setting them as class properties instead of constructor
    arguments.  This is common in clearskies as it helps make easily reusable classes - configuration can
    go in the constructor of the class, allowing the developer to directly instantiate it, and then the DI system
    will come by afterwards and provide the necessary dependencies.

    After adding InjectableProperties as a parent of your class, you have two ways to specify your dependencies:

     1. By using the classes in the `clearskies.di.inject.*`module.
     2. By directly attaching objects which also use the `InjectableProperties` class.

    The following table shows the dependencies that can be injected as properties via the clearskies.di.inject module:

    | Class                            | Type                                 | Result                                          |
    |----------------------------------|--------------------------------------|-------------------------------------------------|
    | clearskies.di.inject.ByClass     | N/A                                  | The specified class will be built               |
    | clearskies.di.inject.ByName      | N/A                                  | The specified dependnecy name will be built     |
    | clearskies.di.inject.Cursor      | N/A                                  | The PyMySQL cursor                              |
    | clearskies.di.inject.Di          | N/A                                  | The dependency injection container itself       |
    | clearskies.di.inject.Environment | clearskies.Environment               | The environment helper                          |
    | clearskies.di.inject.InputOutput | clearskies.input_outputs.InputOutput | The InputOutput object for the current request  |
    | clearskies.di.inject.Now         | datetime.datetime                    | The current time (no timezone)                  |
    | clearskies.di.inject.Requests    | requests.Session                     | A requests session                              |
    | clearskies.di.inject.Utcnow      | datetime.datetime                    | The current time (tzinfo=datetime.timezone.utc) |

    Note: now/utcnow are not cached, so you'll get the current time everytime you get a value out of the class property,
    unless a specific time has been set on the dependency injection container.

    Here's an example:

    ```python
    import clearskies
    import time
    from clearskies import parameters_to_properties


    class MyOtherThing(clearskies.di.InjectableProperties):
        now = clearskies.di.inject.Now()


    class ReusableClass(clearskies.Configurable, clearskies.di.InjectableProperties):
        my_int = clearskies.configs.Integer(required=True)
        some_number = clearskies.di.inject.ByName("some_number")
        my_other_thing = clearskies.di.inject.ByClass(MyOtherThing)

        @parameters_to_properties
        def __init__(self, my_int: int):
            self.finalize_and_validate_configuration()

        def my_value(self) -> int:
            return self.my_int * self.some_number


    class MyClass(clearskies.di.InjectableProperties):
        reusable = ReusableClass(5)


    class MyOtherClass(clearskies.di.InjectableProperties):
        reusable = ReusableClass(10)


    di = clearskies.di.Di(
        bindings={
            "some_number": 10,
        }
    )

    my_class = di.build(MyClass)
    print(my_class.reusable.my_value())  # prints 50

    my_other_class = di.build(MyOtherClass)
    print(my_other_class.reusable.my_value())  # prints 100

    start = my_class.reusable.my_other_thing.now
    time.sleep(1)
    stop = my_class.reusable.my_other_thing.now
    print((stop - start).seconds)  # prints 1
    ```
    """

    _injectables_loaded: str = ""

    @classmethod
    def injectable_properties(cls, di: Di):
        # you would think that I would be able to just use a simple true/false flag attached to the class,
        # but I'm having this weird issue where (when I tried that) the flag was being shared between classes.
        # It shouldn't happen like that, but it is, so there is probably something subtle going on that I
        # haven't figured out yet, but this also works identitally, so :shrug:.
        # Also, keep track of the id of DI.  We use class level caching but tests often use multiple DI containers
        # in one run, which means that we need to re-inject dependencies if we get a new DI container
        cache_name = str(cls) + str(id(di))
        if cache_name == cls._injectables_loaded:
            return

        injectable_descriptors: list[Any] = []
        injectable_properties: list[Self] = []
        for attribute_name in dir(cls):
            # Per the docs above, we want to inject properties for one of two things: the injectables from clearskies.di.inject,
            # and any object that itself extends this class.  This is mildly tricky because the injectables are descriptors, and
            # so we get them using getattr on the class, while if it's not a descriptor, then we want to use getattr on self.
            # The important part here is that we modify descriptors at the class level, so the actual injected values have to
            # be stored in self, and not in the descriptor object.  When it's not a descriptor, then we can modify the object
            # directly (since we're operating at the object level, not class level).  Either way, while we go, let's keep track
            # of what our dependencies are and which ones are cached, so we only have to list the objects attributes the first time.
            attribute = getattr(cls, attribute_name)

            if di.has_class_override(attribute.__class__):
                setattr(cls, attribute_name, di.get_override_by_class(attribute))
                continue

            if issubclass(attribute.__class__, Injectable):
                attribute.set_di(di)
                continue

            if hasattr(attribute, "injectable_properties"):
                attribute.injectable_properties(di)

        cls._injectables_loaded = cache_name



from . import (
    authentication,
    autodoc,
    backends,
    columns,
    configs,
    contexts,
    di,
    endpoints,
    exceptions,
    functional,
    query,
    # secrets,
    security_headers,
    typing,
    validators,
)
from . import parameters_to_properties as parameters_to_properties_module
from .column import Column
from .configurable import Configurable
from .end import End  # type: ignore
from .endpoint import Endpoint
from .endpoint_group import EndpointGroup
from .schema import Schema

parameters_to_properties = parameters_to_properties_module.parameters_to_properties

from .action import Action
from .environment import Environment
from .model import Model
from .security_header import SecurityHeader
from .validator import Validator

__all__ = [
    "Action",
    "Authentication",
    "autodoc",
    "backends",
    "Column",
    "columns",
    "configs",
    "Configurable",
    "contexts",
    "di",
    "End",
    "Endpoint",
    "EndpointGroup",
    "endpoints",
    "Environment",
    "exceptions",
    "functional",
    "Model",
    "parameters_to_properties",
    "Schema",
    "typing",
    "Validator",
    "query",
    # "secrets",
    "SecurityHeader",
    "security_headers",
    "validators",
]



from __future__ import annotations

import inspect
from collections import OrderedDict
from typing import TYPE_CHECKING, Any, Callable

import clearskies.configs
import clearskies.exceptions
from clearskies import authentication, autodoc, typing
from clearskies.endpoint import Endpoint
from clearskies.functional import string
from clearskies.input_outputs import InputOutput

if TYPE_CHECKING:
    from clearskies import Schema, SecurityHeader
    from clearskies.column import Column
    from clearskies.model import Model


class List(Endpoint):
    """
    Create a list endpoint that fetches and returns records to the end client.

    A list endpoint has four required parameters:

    | Name                       | Value                                                                                 |
    |----------------------------|---------------------------------------------------------------------------------------|
    | `model_class`              | The model class for the endpoint to use to find and return records.                   |
    | `readable_column_names`    | A list of columns from the model class that the endpoint should return to the client. |
    | `sortable_column_names`    | A list of columns that the client is allowed to sort by.                              |
    | `default_sort_column_name` | The default column to sort by.                                                        |

    Here's a basic working example:

    ```python
    import clearskies

    class User(clearskies.Model):
        id_column_name = "id"
        backend = clearskies.backends.MemoryBackend()
        id = clearskies.columns.Uuid()
        name = clearskies.columns.String()


    list_users = clearskies.endpoints.List(
        model_class=User,
        readable_column_names=["id", "name"],
        sortable_column_names=["id", "name"],
        default_sort_column_name="name",
    )

    wsgi = clearskies.contexts.WsgiRef(
        list_users,
        classes=[User],
        bindings={
            "memory_backend_default_data": [
                {
                    "model_class": User,
                    "records": [
                        {"id": "1-2-3-4", "name": "Bob"},
                        {"id": "1-2-3-5", "name": "Jane"},
                        {"id": "1-2-3-6", "name": "Greg"},
                    ],
                },
            ]
        },
    )
    wsgi()
    ```

    You can then fetch your records:

    ```bash
    $ curl 'http://localhost:8080/' | jq
    {
        "status": "success",
        "error": "",
        "data": [
            {"id": "1-2-3-4", "name": "Bob"},
            {"id": "1-2-3-6", "name": "Greg"},
            {"id": "1-2-3-5", "name": "Jane"},
        ],
        "pagination": {
            "number_results": 3,
            "limit": 50,
            "next_page": {}
        },
        "input_errors": {}
    }
    ```

    Pagination can be set via query parameters or the JSON body:

    ```bash
    $ curl 'http://localhost:8080/?sort=name&direction=desc&limit=2' | jq
    {
        "status": "success",
        "error": "",
        "data": [
            {"id": "1-2-3-5", "name": "Jane"},
            {"id": "1-2-3-6", "name": "Greg"},
        ],
        "pagination": {
            "number_results": 3,
            "limit": 2,
            "next_page": {"start": 2}
        },
        "input_errors": {}
    }
    ```

    In the response, '.pagination.next_page` is a dictionary that returns the query parameters to set in order to fetch the next page of results.
    Note that the pagination method depends on the backend.  The memory backend supports pagination via start/limit, while other backends may
    support alternate pagination schemes.  Clearskies automatically handles the difference, so it's important to use `.pagination.next_page` to fetch
    the next page of results.

    Use `where`, `joins`, and `group_by` to automatically adjust the query used by the list endpoint.  In particular, where is a list of either
    conditions (as a string) or a callable that can modify the query directly via the model class.  For example:

    ```python
    list_users = clearskies.endpoints.List(
        model_class=User,
        readable_column_names=["id", "name"],
        sortable_column_names=["id", "name"],
        default_sort_column_name="name",
        where=[User.name.equals("Jane")],  # equivalent: where=["name=Jane"]
    )
    ```

    With the above definition, the list endpoint will only ever return records with a name of "Jane".  The following uses standard dependency
    injection rules to execute a similar filter based on arbitrary logic required:

    ```python
    import datetime

    list_users = clearskies.endpoints.List(
        model_class=User,
        readable_column_names=["id", "name"],
        sortable_column_names=["id", "name"],
        default_sort_column_name="name",
        where=[
            lambda model, now: model.where("name=Jane")
            if now > datetime.datetime(2025, 1, 1)
            else model
        ],
    )
    ```

    As shown in the above example, a function called in this way can request additional dependencies as needed, per the standard dependency rules.
    The function needs to return the adjusted model object, which is usually as simple as returning the result of `model.where(?)`.  While the
    above example uses a lambda function, of course you can attach any other kind of callable - a function, a method of a class, etc...
    """

    """
    The default column to sort by.
    """
    default_sort_column_name = clearskies.configs.ModelColumn("model_class")

    """
    The default sort direction (ASC or DESC).
    """
    default_sort_direction = clearskies.configs.Select(["ASC", "DESC"], default="ASC")

    """
    The number of records returned if the client doesn't specify a different number of records (default: 50).
    """
    default_limit = clearskies.configs.Integer(default=50)

    """
    The maximum number of records the client is allowed to request (0 == no limit)
    """
    maximum_limit = clearskies.configs.Integer(default=200)

    """
    A column to group by.
    """
    group_by_column_name = clearskies.configs.ModelColumn("model_class")

    readable_column_names = clearskies.configs.ReadableModelColumns("model_class")
    sortable_column_names = clearskies.configs.ReadableModelColumns("model_class", allow_relationship_references=True)
    searchable_column_names = clearskies.configs.SearchableModelColumns(
        "model_class", allow_relationship_references=True
    )

    @clearskies.parameters_to_properties.parameters_to_properties
    def __init__(
        self,
        model_class: type[Model],
        readable_column_names: list[str],
        sortable_column_names: list[str],
        default_sort_column_name: str | None,
        default_sort_direction: str = "ASC",
        default_limit: int = 50,
        maximum_limit: int = 200,
        where: typing.condition | list[typing.condition] = [],
        joins: typing.join | list[typing.join] = [],
        url: str = "",
        request_methods: list[str] = ["GET"],
        response_headers: list[str | Callable[..., list[str]]] = [],
        output_map: Callable[..., dict[str, Any]] | None = None,
        output_schema: Schema | None = None,
        column_overrides: dict[str, Column] = {},
        group_by_column_name: str = "",
        internal_casing: str = "snake_case",
        external_casing: str = "snake_case",
        security_headers: list[SecurityHeader] = [],
        description: str = "",
        authentication: authentication.Authentication = authentication.Public(),
        authorization: authentication.Authorization = authentication.Authorization(),
    ):
        # we need to call the parent but don't have to pass along any of our kwargs.  They are all optional in our parent, and our parent class
        # just stores them in parameters, which we have already done.  However, the parent does do some extra initialization stuff that we need,
        # which is why we have to call the parent.
        super().__init__()

    @property
    def searchable_columns(self) -> dict[str, Column]:
        if self._searchable_columns is None:
            self._searchable_columns = {name: self.columns[name] for name in self.searchable_column_names}
        return self._searchable_columns

    @property
    def sortable_columns(self) -> dict[str, Column]:
        if self._sortable_columns is None:
            self._sortable_columns = {name: self.columns[name] for name in self.sortable_column_names}
        return self._sortable_columns

    @property
    def allowed_request_keys(self) -> list[str]:
        return [*["sort", "direction", "limit"], *self.searchable_column_names]

    @property
    def internal_request_keys(self) -> list[str]:
        return ["sort", "direction", "limit"]

    def handle(self, input_output: InputOutput):
        model = self.fetch_model_with_base_query(input_output)
        if not input_output.request_data and input_output.has_body():
            raise clearskies.exceptions.ClientError("Request body was not valid JSON")
        if input_output.request_data and not isinstance(input_output.request_data, dict):
            raise clearskies.exceptions.ClientError("When present, request body must be a JSON dictionary")
        request_data = self.map_input_to_internal_names(input_output.request_data)  # type: ignore
        query_parameters = self.map_input_to_internal_names(input_output.query_parameters)
        pagination_data = {}
        for key in model.allowed_pagination_keys():
            if key in request_data and key in query_parameters:
                original_name = self.auto_case_internal_column_name(key)
                raise clearskies.exceptions.ClientError(
                    f"Ambiguous request: key '{original_name}' is present in both the JSON body and URL data"
                )
            if key in request_data:
                pagination_data[key] = request_data[key]
                del request_data[key]
            if key in query_parameters:
                pagination_data[key] = query_parameters[key]
                del query_parameters[key]
        if request_data or query_parameters or pagination_data:
            self.check_request_data(request_data, query_parameters, pagination_data)
            model = self.configure_model_from_request_data(model, request_data, query_parameters, pagination_data)
        if not model.get_query().limit:
            model = model.limit(self.default_limit)
        if not model.get_query().sorts and self.default_sort_column_name:
            model = model.sort_by(
                self.default_sort_column_name,
                self.default_sort_direction,
                model.destination_name(),
            )
        if self.group_by_column_name:
            model = model.group_by(self.group_by_column_name)

        return self.success(
            input_output,
            [self.model_as_json(record, input_output) for record in model],
            number_results=len(model) if model.backend.can_count else None,
            limit=model.get_query().limit,
            next_page=model.next_page_data(),
        )

    def configure_model_from_request_data(
        self,
        model: Model,
        request_data: dict[str, Any],
        query_parameters: dict[str, Any],
        pagination_data: dict[str, Any],
    ) -> Model:
        limit = int(self.from_either(request_data, query_parameters, "limit", default=self.default_limit))
        model = model.limit(limit)
        if pagination_data:
            model = model.pagination(**pagination_data)
        sort = self.from_either(request_data, query_parameters, "sort")
        direction = self.from_either(request_data, query_parameters, "direction")
        if sort and direction:
            model = self.add_join(sort, model)
            [sort_column, sort_table] = self.resolve_references_for_query(sort)
            model = model.sort_by(sort_column, direction, sort_table)  # type: ignore

        return model

    def map_input_to_internal_names(self, data: dict[str, Any]) -> dict[str, Any]:
        if not data:
            return {}
        internal_request_keys = [*self.internal_request_keys, *self.model.allowed_pagination_keys()]
        for key in internal_request_keys:
            mapped_key = self.auto_case_internal_column_name(key)
            if mapped_key != key and mapped_key in data:
                data[key] = data[mapped_key]
                del data[mapped_key]
        # any non-internal fields are assumed to be column names and need to go
        # through the full mapping
        for key in set(self.allowed_request_keys) - set(internal_request_keys):
            mapped_key = self.auto_case_column_name(key, True)
            if mapped_key != key and mapped_key in data:
                data[key] = data[mapped_key]
                del data[mapped_key]

        # finally, if we have a sort key set then convert the value to the properly cased column name
        if "sort" in data:
            # we can't just take the sort value and convert it to internal casing because camel/title case
            # to snake_case can be ambiguous (while snake_case to camel/title is not)
            sort_column_map = {}
            for internal_name in self.sortable_column_names:
                external_name = self.auto_case_column_name(internal_name, True)
                sort_column_map[external_name] = internal_name
            # sometimes the sort may be a list of directives
            if isinstance(data["sort"], list):
                for index, sort_entry in enumerate(data["sort"]):
                    if "column" not in sort_entry:
                        continue
                    if sort_entry["column"] in sort_column_map:
                        sort_entry["column"] = sort_column_map[sort_entry["column"]]
            else:
                if data["sort"] in sort_column_map:
                    data["sort"] = sort_column_map[data["sort"]]

        return data

    def check_request_data(
        self, request_data: dict[str, Any], query_parameters: dict[str, Any], pagination_data: dict[str, Any]
    ) -> None:
        if pagination_data:
            error = self.model.validate_pagination_data(pagination_data, self.auto_case_internal_column_name)
            if error:
                raise clearskies.exceptions.ClientError(error)
        for key in request_data.keys():
            if key not in self.allowed_request_keys:
                raise clearskies.exceptions.ClientError(f"Invalid request parameter found in request body: '{key}'")
        for key in query_parameters.keys():
            if key not in self.allowed_request_keys:
                raise clearskies.exceptions.ClientError(f"Invalid request parameter found in URL data: '{key}'")
            if key in request_data:
                raise clearskies.exceptions.ClientError(
                    f"Ambiguous request: '{key}' was found in both the request body and URL data"
                )
        self.validate_limit(request_data, query_parameters)
        sort = self.from_either(request_data, query_parameters, "sort")
        direction = self.from_either(request_data, query_parameters, "direction")
        if sort and type(sort) != str:
            raise clearskies.exceptions.ClientError("Invalid request: 'sort' should be a string")
        if direction and type(direction) != str:
            raise clearskies.exceptions.ClientError("Invalid request: 'direction' should be a string")
        if sort or direction:
            if (sort and not direction) or (direction and not sort):
                raise clearskies.exceptions.ClientError(
                    "You must specify 'sort' and 'direction' together in the request - not just one of them"
                )
            if sort not in self.sortable_column_names:
                raise clearskies.exceptions.ClientError(f"Invalid request: invalid sort column")
            if direction.lower() not in ["asc", "desc"]:
                raise clearskies.exceptions.ClientError("Invalid request: direction must be 'asc' or 'desc'")
        self.check_search_in_request_data(request_data, query_parameters)

    def validate_limit(self, request_data: dict[str, Any], query_parameters: dict[str, Any]) -> None:
        limit = self.from_either(request_data, query_parameters, "limit")
        if limit is not None and type(limit) != int and type(limit) != float and type(limit) != str:
            raise clearskies.exceptions.ClientError("Invalid request: 'limit' should be an integer")
        if limit:
            try:
                limit = int(limit)
            except ValueError:
                raise clearskies.exceptions.ClientError("Invalid request: 'limit' should be an integer")
        if limit:
            if limit > self.maximum_limit:
                raise clearskies.exceptions.ClientError(f"Invalid request: 'limit' must be at most {self.max_limit}")
            if limit < 0:
                raise clearskies.exceptions.ClientError(f"Invalid request: 'limit' must be positive")

    def check_search_in_request_data(self, request_data: dict[str, Any], query_parameters: dict[str, Any]):
        return None

    def unpack_column_name_with_relationship(self, column_name: str) -> list[str]:
        if "." not in column_name:
            return ["", column_name]
        return column_name.split(".", 1)

    def resolve_references_for_query(self, column_name: str) -> list[str | None]:
        """
        Take the column name and returns the name and table.

        If it's just a column name, we assume the table is the table for our model class.
        If it's something like `belongs_to_column.column_name`, then it will find the appropriate
        table reference.
        """
        if not column_name:
            return [None, None]
        [relationship_column_name, column_name] = self.unpack_column_name_with_relationship(column_name)
        if not relationship_column_name:
            return [self.model.destination_name(), column_name]

        return [self.columns[relationship_column_name].join_table_alias(), column_name]

    def add_join(self, column_name: str, model: Model) -> Model:
        """
        Add a join to the query for the given column name in the case where it references a column in a belongs to.

        If column_name is something like `belongs_to_column.column_name`, this will add have the belongs to column
        add it's typical join condition, so that further sorting/searching can work.

        If column_name is empty, or doesn't contain a period, then this does nothing.
        """
        if not column_name:
            return model
        [relationship_column_name, column_name] = self.unpack_column_name_with_relationship(column_name)
        if not relationship_column_name:
            return model
        return self.columns[relationship_column_name].add_join(model)

    def from_either(self, request_data, query_parameters, key, default=None, ignore_none=True):
        """Return the key from either object.  Assumes it is not present in both."""
        if key in request_data:
            if request_data[key] is not None or not ignore_none:
                return request_data[key]
        if key in query_parameters:
            if query_parameters[key] is not None or not ignore_none:
                return query_parameters[key]
        return default

    def documentation(self) -> list[autodoc.request.Request]:
        nice_model = string.camel_case_to_words(self.model_class.__name__)
        schema_model_name = string.camel_case_to_snake_case(self.model_class.__name__)
        data_schema = self.documentation_data_schema()

        authentication = self.authentication
        standard_error_responses = []
        if not getattr(authentication, "is_public", False):
            standard_error_responses.append(self.documentation_access_denied_response())
            if getattr(authentication, "can_authorize", False):
                standard_error_responses.append(self.documentation_unauthorized_response())

        return [
            autodoc.request.Request(
                f"Fetch the list of current {nice_model} records",
                [
                    self.documentation_success_response(
                        autodoc.schema.Array(
                            self.auto_case_internal_column_name("data"),
                            autodoc.schema.Object(nice_model, children=data_schema, model_name=schema_model_name),
                        ),
                        description=f"The matching {nice_model} records",
                        include_pagination=True,
                    ),
                    *standard_error_responses,
                    self.documentation_generic_error_response(),
                ],
                relative_path=self.url,
                request_methods=self.request_methods,
                parameters=self.documentation_request_parameters(),
                root_properties={
                    "security": self.documentation_request_security(),
                },
            ),
        ]

    def documentation_request_parameters(self) -> list[autodoc.request.Parameter]:
        return [
            *self.documentation_url_pagination_parameters(),
            *self.documentation_url_sort_parameters(),
            *self.documentation_url_search_parameters(),
            *self.documentation_json_search_parameters(),
        ]

    def documentation_models(self) -> dict[str, autodoc.schema.Schema]:
        schema_model_name = string.camel_case_to_snake_case(self.model_class.__name__)

        return {
            schema_model_name: autodoc.schema.Object(
                self.auto_case_internal_column_name("data"),
                children=self.documentation_data_schema(),
            ),
        }

    def documentation_url_pagination_parameters(self) -> list[autodoc.request.Parameter]:
        url_parameters = [
            autodoc.request.URLParameter(
                autodoc.schema.Integer(self.auto_case_internal_column_name("limit")),
                description="The number of records to return",
            ),
        ]

        for parameter in self.model.documentation_pagination_parameters(self.auto_case_internal_column_name):
            (schema, description) = parameter
            url_parameters.append(autodoc.request.URLParameter(schema, description=description))

        return url_parameters  # type: ignore

    def documentation_url_sort_parameters(self) -> list[autodoc.request.Parameter]:
        sort_columns = [self.auto_case_column_name(internal_name, True) for internal_name in self.sortable_column_names]
        directions = [self.auto_case_column_name(internal_name, True) for internal_name in ["asc", "desc"]]

        return [
            autodoc.request.URLParameter(
                autodoc.schema.Enum(
                    self.auto_case_internal_column_name("sort"),
                    sort_columns,
                    autodoc.schema.String(self.auto_case_internal_column_name("sort")),
                    example=self.auto_case_column_name("name", True),
                ),
                description=f"Column to sort by",
            ),
            autodoc.request.URLParameter(
                autodoc.schema.Enum(
                    self.auto_case_internal_column_name("direction"),
                    directions,
                    autodoc.schema.String(self.auto_case_internal_column_name("direction")),
                    example=self.auto_case_column_name("asc", True),
                ),
                description=f"Direction to sort",
            ),
        ]

    def documentation_json_pagination_parameters(self) -> list[autodoc.request.Parameter]:
        json_parameters = [
            autodoc.request.JSONBody(
                autodoc.schema.Integer(self.auto_case_internal_column_name("limit")),
                description="The number of records to return",
            ),
        ]

        for parameter in self.model.documentation_pagination_parameters(self.auto_case_internal_column_name):
            (schema, description) = parameter
            json_parameters.append(autodoc.request.JSONBody(schema, description=description))

        return json_parameters  # type: ignore

    def documentation_json_sort_parameters(self) -> list[autodoc.request.Parameter]:
        sort_columns = [self.auto_case_column_name(internal_name, True) for internal_name in self.sortable_column_names]
        directions = [self.auto_case_column_name(internal_name, True) for internal_name in ["asc", "desc"]]

        return [
            autodoc.request.JSONBody(
                autodoc.schema.Enum(
                    self.auto_case_internal_column_name("sort"),
                    sort_columns,
                    autodoc.schema.String(self.auto_case_internal_column_name("sort")),
                    example=self.auto_case_column_name("name", True),
                ),
                description=f"Column to sort by",
            ),
            autodoc.request.JSONBody(
                autodoc.schema.Enum(
                    self.auto_case_internal_column_name("direction"),
                    directions,
                    autodoc.schema.String(self.auto_case_internal_column_name("direction")),
                    example=self.auto_case_column_name("asc", True),
                ),
                description=f"Direction to sort",
            ),
        ]

    def documentation_url_search_parameters(self) -> list[autodoc.request.Parameter]:
        return []

    def documentation_json_search_parameters(self) -> list[autodoc.request.Parameter]:
        return []



"""
This module helps classes declare their configuration parameters via properties.

To use it, the class needs to include the clearskies.configs.Confirgurable in its parent chain and
then create properties as needed using the various classes in the clearskies.configs module.  Each
class represents a specific "kind" of configuration, has typing declared to help while writing code,
and runtime checks to verify configs while the application is (preferably) booting.

Each config accepts a `required` and `default` kwarg to assist with validation/construction of
the configuration

Data is stored in the `_config` property on the instance.

Usage:

```python
from clearskies import configs


class MyConfigurableClass(configs.Configurable):
    name = configs.String()
    age = configs.Integer(required=True)
    property_with_default = configs.String(default="some value")

    def __init__(self, name, age, optional=None):
        self.name = name
        self.age = age

        # always call this after saving the confiuration values to the properties.
        # It will fill in default values for any properties that have a default and
        # are none, and it will raise a ValueError if there is a required property
        # that does not have a value.
        self.finalize_and_validate_configuration()


configured_thingie = MyConfigurableClass("Bob", 18)
print(configured_thingie.age)  # prints: 18
print(configured_thingie.property_with_default)  # prints: some value

invalid_thingie = MyConfigurableClass(18, 20)  # raises a TypeError

also_invalid = MyConfigurableClass("", 18)  # raises a ValueError
```

Finally, parameters_to_properties is a decorator that will take any parameters passed into the
decorated function and assign them as instance properties.  You can use this to skip some code,
especially if you have a lot of configuration parameters.  In the above example, you could simplify
it as:

```python
from clearskies import configs


class MyConfigurableClass(configs.Configurable):
    name = configs.String()
    age = configs.Integer(required=True)
    property_with_default = configs.String(default="some value")

    @clearskies.parameters_to_properties()
    def __init__(self, name: str, age: int, optional: string = None):
        self.finalize_and_validate_configuration()
```

"""

import inspect

from .actions import Actions
from .any import Any
from .any_dict import AnyDict
from .any_dict_or_callable import AnyDictOrCallable
from .authentication import Authentication
from .authorization import Authorization
from .boolean import Boolean
from .boolean_or_callable import BooleanOrCallable
from .callable_config import Callable
from .columns import Columns
from .conditions import Conditions
from .config import Config
from .datetime import Datetime
from .datetime_or_callable import DatetimeOrCallable
from .endpoint import Endpoint
from .float import Float
from .float_or_callable import FloatOrCallable
from .integer import Integer
from .integer_or_callable import IntegerOrCallable
from .joins import Joins
from .list_any_dict import ListAnyDict
from .list_any_dict_or_callable import ListAnyDictOrCallable
from .model_class import ModelClass
from .model_column import ModelColumn
from .model_columns import ModelColumns
from .model_destination_name import ModelDestinationName
from .model_to_id_column import ModelToIdColumn
from .readable_model_column import ReadableModelColumn
from .readable_model_columns import ReadableModelColumns
from .schema import Schema
from .searchable_model_columns import SearchableModelColumns
from .security_headers import SecurityHeaders
from .select import Select
from .select_list import SelectList
from .string import String
from .string_dict import StringDict
from .string_list import StringList
from .string_list_or_callable import StringListOrCallable
from .string_or_callable import StringOrCallable
from .timedelta import Timedelta
from .timezone import Timezone
from .url import Url
from .validators import Validators
from .writeable_model_column import WriteableModelColumn
from .writeable_model_columns import WriteableModelColumns

__all__ = [
    "Actions",
    "Any",
    "AnyDict",
    "AnyDictOrCallable",
    "Authentication",
    "Authorization",
    "Boolean",
    "BooleanOrCallable",
    "Callable",
    "Columns",
    "Conditions",
    "Config",
    "Datetime",
    "DatetimeOrCallable",
    "Float",
    "FloatOrCallable",
    "Joins",
    "Integer",
    "IntegerOrCallable",
    "ListAnyDict",
    "ListAnyDictOrCallable",
    "ModelClass",
    "ModelColumn",
    "ModelColumns",
    "ModelToIdColumn",
    "ModelDestinationName",
    "ReadableModelColumn",
    "ReadableModelColumns",
    "Schema",
    "SearchableModelColumns",
    "SecurityHeaders",
    "Select",
    "SelectList",
    "String",
    "StringDict",
    "StringList",
    "StringListOrCallable",
    "StringOrCallable",
    "Timedelta",
    "Timezone",
    "Url",
    "Validators",
    "WriteableModelColumn",
    "WriteableModelColumns",
]



from clearskies.contexts.cli import Cli
from clearskies.contexts.context import Context
from clearskies.contexts.wsgi import Wsgi
from clearskies.contexts.wsgi_ref import WsgiRef

__all__ = [
    "Cli",
    "Context",
    "Wsgi",
    "WsgiRef",
]



class Condition:
    """
    Parses a condition string, e.g. "column=value" or "table.column<=other_value".

    Allowed operators: ["<=>", "!=", "<=", ">=", ">", "<", "=", "in", "is not null", "is null", "is not", "is", "like"]

    NOTE: Not all backends support all operators, so make sure the condition you are building works for your backend

    This is safe to use with untrusted input because it expects a stringent and easy-to-verify format.  The incoming
    string must be one of these patterns:

     1. [column_name][operator][value]
     2. [table_name].[column_name][operator][value]

    SQL-like syntax is allowed, so:

     1. Spaces are optionally allowed around the operator.
     2. Backticks are optionally allowd around the table/column name.
     3. Single quotes are optionally allowed around the values.
     4. operators are case-insensitive.

    In the case of an IN operator, the parser expects a series of comma separated values enclosed in parenthesis,
    with each value optionally enclosed in single quotes.  This parsing is very simple and there is not currently a way
    to escape commas or single quotes.

    NOTE: operators (when they are english words, of course) are always output in all upper-case.

    Some examples:

    ```python
    condition = Condition("id=asdf-qwerty")  # note: same results for: Condition("id = asdf-qwerty")
    print(condition.table_name)  # prints ''
    print(condition.column_name)  # prints 'id'
    print(condition.operator)  # prints '='
    print(condition.values)  # prints ['asdf-qwerty']
    print(condition.parsed)  # prints 'id=%s'

    condition = Condition("orders.status_id in ('ACTIVE', 'PENDING')")
    print(condition.table_name)  # prints 'orders'
    print(condition.column_name)  # prints 'status_id'
    print(condition.operator)  # prints 'IN'
    print(condition.values)  # prints ['ACTIVE', 'PENDING']
    print(condition.parsed)  # prints 'status_id IN (%s, %s)'
    ```
    """

    """
    The name of the table this condition is searching on (if there is one).
    """
    table_name: str = ""

    """
    The name of the column the condition is searching.
    """
    column_name: str = ""

    """
    The operator we are searching with (e.g. '=', '<=', etc...)
    """
    operator: str = ""

    """
    The values the condition is searching for.

    Note this is always a list, although most of the time there is only one value in the list.  Multiple values
    are only present when searching with the IN operator.
    """
    values: list[str] = []

    """
    An SQL-ready string
    """
    parsed: str = ""

    """
    The original condition string
    """
    _raw_condition: str = ""

    """
    The list of operators we can match

    Note: the order is very important because this list is used to find the operator in the condition string.
    As a result, the order of the operators in this list is important.  The condition matching algorithm used
    below will select whichever operator matches earlier in the string, but there are some operators that
    start with the same characters: '<=>' and '<=', as well as 'is', 'is null', 'is not', etc...  This leaves
    room for ambiguity since all of these operators will match at the same location.  In the event of a "tie" the
    algorithm gives preference to the first matching operator.  Therefore, for ambiguous operators, we put the
    longer one first, which means it matches first, and so a condition with a '<=>' operator won't accidentally
    match to the '<=' operator.
    """
    operators: list[str] = [
        "<=>",
        "!=",
        "<=",
        ">=",
        ">",
        "<",
        "=",
        "in",
        "is not null",
        "is null",
        "is not",
        "is",
        "like",
    ]

    operator_lengths: dict[str, int] = {
        "<=>": 3,
        "<=": 2,
        ">=": 2,
        "!=": 2,
        ">": 1,
        "<": 1,
        "=": 1,
        "in": 4,
        "is not null": 12,
        "is null": 8,
        "is not": 8,
        "is": 4,
        "like": 6,
    }

    # some operators require spaces around them
    operators_for_matching: dict[str, str] = {
        "like": " like ",
        "in": " in ",
        "is not null": " is not null",
        "is null": " is null",
        "is": " is ",
        "is not": " is not ",
    }

    operators_with_simple_placeholders: dict[str, bool] = {
        "<=>": True,
        "<=": True,
        ">=": True,
        "!=": True,
        "=": True,
        "<": True,
        ">": True,
    }

    operators_without_placeholders: dict[str, bool] = {
        "is not null": True,
        "is null": True,
    }

    def __init__(self, condition: str):
        self._raw_condition = condition
        lowercase_condition = condition.lower()
        self.operator = ""
        matching_index = len(condition)
        # figure out which operator comes earliest in the string: make sure and check all so we match the
        # earliest operator so we don't get unpredictable results for things like 'age=name<=5'.  We want
        # our operator to **ALWAYS** match whatever comes first in the condition string.
        for operator in self.operators:
            try:
                operator_for_match = self.operators_for_matching.get(operator, operator)
                index = lowercase_condition.index(operator_for_match)
            except ValueError:
                continue
            if index < matching_index:
                matching_index = index
                self.operator = operator

        if not self.operator:
            raise ValueError(f"No supported operators found in condition {condition}")

        self.column_name = condition[:matching_index].strip().replace("`", "")
        value = condition[matching_index + self.operator_lengths[self.operator] :].strip()
        if value and (value[0] == "'" and value[-1] == "'"):
            value = value.strip("'")
        self.values = self._parse_condition_list(value) if self.operator == "in" else [value]
        self.table_name = ""
        if "." in self.column_name:
            [self.table_name, self.column_name] = self.column_name.split(".")
        column_for_parsed = f"{self.table_name}.{self.column_name}" if self.table_name else self.column_name

        if self.operator in self.operators_without_placeholders:
            self.values = []

        self.operator = self.operator.upper()
        self.parsed = self._with_placeholders(
            column_for_parsed, self.operator, self.values, escape=False if self.table_name else True
        )

    def _parse_condition_list(self, value):
        if value[0] != "(" and value[-1] != ")":
            raise ValueError(f"Invalid search value {value} for condition.  For IN operator use `IN (value1,value2)`")

        # note: this is not very smart and will mess things up if there are single quotes/commas in the data
        return list(map(lambda value: value.strip().strip("'"), value[1:-1].split(",")))

    def _with_placeholders(self, column, operator, values, escape=True, escape_character="`"):
        quote = escape_character if escape else ""
        column = column.replace("`", "")
        upper_case_operator = operator.upper()
        lower_case_operator = operator.lower()
        if lower_case_operator in self.operators_with_simple_placeholders:
            return f"{quote}{column}{quote}{upper_case_operator}%s"
        if lower_case_operator in self.operators_without_placeholders:
            return f"{quote}{column}{quote} {upper_case_operator}"
        if lower_case_operator == "is" or lower_case_operator == "is not" or lower_case_operator == "like":
            return f"{quote}{column}{quote} {upper_case_operator} %s"

        # the only thing left is "in" which has a variable number of placeholders
        return f"{quote}{column}{quote} IN (" + ", ".join(["%s" for i in range(len(values))]) + ")"


class ParsedCondition(Condition):
    def __init__(self, column_name: str, operator: str, values: list[str], table_name: str = ""):
        self.column_name = column_name
        if operator not in self.operators:
            raise ValueError(f"Unknown operator '{operator}'")
        self.operator = operator
        self.values = values
        self.table_name = table_name
        column_for_parsed = f"{self.table_name}.{self.column_name}" if self.table_name else self.column_name
        self.parsed = self._with_placeholders(
            column_for_parsed, self.operator, self.values, escape=False if self.table_name else True
        )
        self._raw_condition = self.parsed
